"""Class definition for a matrix representation of a Hamiltonian."""

from abc import abstractmethod

import jax.numpy as jnp
from jax import vmap

from paraqeet.model.drive import Drive
from paraqeet.optimizable import Optimizable
from paraqeet.quantity import Array, Quantity


class Hamiltonian(Optimizable):
    """Class definition for a matrix representation of a Hamiltonian.

    Implementations can contain subsystems, couplings, and drive lines
    and have to take care of frame transformations. Derived classes need to
    implement the functions get_matrix, gradient, dimension, get_collapse_ops,
    get_value_at_timestep

    Parameters
    ----------
    drives : list[Drive]
        List of time-dependent drives.

    """

    _drives: list[Drive]

    def __init__(self, drives: list[Drive] | None = None):
        self._drives = [d for d in drives if d is not None] if drives else []

    @abstractmethod
    def dimension(self) -> int:
        """Return the dimension of the Hilbert space of this Hamiltonian.

        Returns
        -------
        int
            Returns the dimension of the Hilbert space of this Hamiltonian.

        """
        pass

    def get_value(self, times: Array) -> Array:
        """Return the matrix representation of the Hamiltonian.

        The default implementation calls get_value_at_timestep for each time step.
        Subclasses can override this function for a more efficient
        implementation.

        Parameters
        ----------
        times: Array
            Vector of time samples.

        Returns
        -------
        Array
            Hamiltonian of shape [t, n, n]  with 't' as time and 'n' as the
            Hilbert space dimension.

        """
        # Ignoring mypy due to vmap
        return jnp.array(vmap(self.get_value_at_timestep)(times))  # type: ignore

    @abstractmethod
    def get_value_at_timestep(self, timestep: float) -> Array:
        """Return the matrix representation of the Hamiltonian.

        Parameters
        ----------
        timestep: float
            One time point.

        Returns
        -------
        Array
            Hamiltonian of shape [n, n]  with `n` as the Hilbert space
            dimension.

        """
        pass

    @property
    def drives(self) -> list[Drive]:
        """Return the list of Drives of the system.

        Returns
        -------
        list[Drive]
            Returns a list of time-dependent drives of the system.
        """
        return self._drives

    @drives.setter
    def drives(self, drives: list[Drive]) -> None:
        """Set the drives

        Parameters
        ----------
        drives: list[Drive]
            List of drives to set.
        """
        self._drives = drives

    def _get_drive_parameters(self) -> list[Quantity]:
        """Return the combined list of parameters from all drives.

        Returns
        -------
        list[Quantity]
            Returns the combined list of parameters from all drives.

        """
        params = []
        for d in self._drives:
            params += d.get_parameters()
        return params

    def _get_drive_matrix(self, annihilation_operator: Array, times: Array) -> Array:
        """Return the sum of all drives in matrix form.

        This function can be used be Hamiltonian implementations for
        including the drive. The default implementation calls
        _get_drive_matrix_one_time for each time step. Subclasses can override this
        function for a more efficient implementation.

        Parameters
        ----------
        annihilation_operator : Array
            The annihilation operator.
        times : Array
            Vector of time samples.

        Returns
        -------
        Array
            Returns the sum of all drives in matrix form.

        """
        # ignoring mypy due to vmap
        return vmap(self._get_drive_matrix_at_timestep, in_axes=(None, 0))(annihilation_operator, times)  # type: ignore

    def _get_drive_matrix_at_timestep(self, annihilation_operator: Array, times: float) -> Array:
        """Return the sum of all drives in matrix form.

        This function can be used be Hamiltonian implementations
        for including the drive.

        Parameters
        ----------
        annihilation_operator : Array
            The annihilation operator.
        times: Array
            Vector of time samples.

        Returns
        -------
        Array
            Returns the sum of all drives in matrix form.

        """
        dim = self.dimension()
        mat = jnp.zeros((dim, dim))
        for drive in self._drives:
            mat += drive.get_value_at_timestep(annihilation_operator, times)
        return mat

    def _get_drive_gradients(self, annihilation_operator: Array, times: Array) -> Array:
        """Return the gradients of all drives.

        This function can be used by Hamiltonian implementations
        for including the drive gradients.

        Parameters
        ----------
        annihilation_operator : Array
            The annihilation operator.
        times: Array
            Vector of time samples.

        Returns
        -------
        Array
            Returns the gradients of all drives.

        """
        dim = self.dimension()
        all_grads = jnp.zeros((times.shape[0], 0, dim, dim))
        for drive in self._drives:
            grads = drive.get_gradient(annihilation_operator, times)
            all_grads = jnp.append(all_grads, grads, axis=1)
        return all_grads

    def _get_drive_gradients_at_timestep(self, annihilation_operator: Array, time: float) -> Array:
        """Return the gradients of all drives.

        This function can be used by Hamiltonian implementations
        for including the drive gradients.

        Parameters
        ----------
        annihilation_operator : Array
            The annihilation operator.
        times: Array
            One time stamp.

        Returns
        -------
        Array
            Returns the gradients of all drives.

        """
        dim = self.dimension()
        all_grads = jnp.zeros((0, dim, dim))
        for drive in self._drives:
            grads = drive.get_gradient_at_timestep(annihilation_operator, time)
            all_grads = jnp.append(all_grads, grads, axis=0)
        return all_grads

    @staticmethod
    def _repeat(mat: Array, num: int) -> Array:
        """Repeat the matrix across time steps.

        Utility function that repeats the matrix mat for each timestep
        in the `num` array. Returns an array with shape [t, n, m] where
        't' is the number of time steps and 'mat' is an 'n' times 'm' matrix.

        Parameters
        ----------
        mat: Array
            Matrix for repetition.
        num: int
            Number of repetitions.

        Returns
        -------
        Array
            Repeated matrix for each time step specified.

        """
        return mat.reshape((1,) + mat.shape).repeat(num, axis=0)

    @abstractmethod
    def get_collapseops(self) -> list[tuple[Array, Array]]:
        """
        Return a list tuples of decay rates and collapse operators for each subsystem.

        Returns
        -------
        list[Tuple[Array, Array]]
            List of collapse operators
        """
        pass



"""Test the coupling model."""

import numpy as np
import pytest

from paraqeet.model.coupling import TwoBodyCoupling
from paraqeet.model.transmon import Transmon
from paraqeet.quantity import Quantity

COUPLINGSTR = 25e6 * 2 * np.pi


@pytest.fixture
def transmon_parameters():
    """Return a random parameter object for transmons."""

    class RandomParameters:
        def get(self):
            freq = np.random.uniform(5.5, 6.0) * 1e9 * 2 * np.pi
            anharm = -np.random.uniform(2.0, 2.4) * 1e7 * 2 * np.pi
            return (freq, anharm)

    return RandomParameters()


@pytest.fixture
def transmon(transmon_parameters):
    """Return a transmon created from the given parameters."""

    class CreateTransmon:
        def get(self, dimension):
            freq, anharm = transmon_parameters.get()
            transmon = Transmon(
                dimension=dimension,
                frequency=Quantity(freq, 0.8 * freq, 1.2 * freq),
                anharmonicity=Quantity(anharm, 1.2 * anharm, 0.8 * anharm),
            )
            return transmon

    return CreateTransmon()


@pytest.fixture
def coupling(transmon):
    """Return a coupling generator method."""

    def _method(dim1: int, dim2: int, is_longitudinal: bool, use_rwa: bool = False):
        transmonA = transmon.get(dim1)
        transmonB = transmon.get(dim2)
        coupling = TwoBodyCoupling(
            subsystem_A=transmonA,
            subsystem_B=transmonB,
            is_longitudinal=is_longitudinal,
            use_rwa=use_rwa,
            coefficient=Quantity(COUPLINGSTR, 0.8 * COUPLINGSTR, 1.2 * COUPLINGSTR, "Hz"),
        )

        return coupling

    return _method


def test_get_matrices(coupling):
    """Test the get matrice method workings."""
    for _ in range(10):
        dim1 = np.random.randint(2, 7)
        dim2 = np.random.randint(2, 7)
        dims = [dim1, dim2]

        # Test shape for Longitudinal coupling
        coup = coupling(dim1, dim2, is_longitudinal=True)
        coup_hams = coup.get_couplings()
        for term in coup_hams:
            for i, ops in enumerate(term):
                assert np.shape(ops) == (dims[i], dims[i])

        # Test for RWA
        coup = coupling(dim1, dim2, is_longitudinal=False, use_rwa=True)
        coup_hams = coup.get_couplings()
        for term in coup_hams:
            for i, ops in enumerate(term):
                assert np.shape(ops) == (dims[i], dims[i])

        # Test shape for Transverse coupling
        coup = coupling(dim1, dim2, is_longitudinal=False)
        coup_hams = coup.get_couplings()
        for term in coup_hams:
            for i, ops in enumerate(term):
                assert np.shape(ops) == (dims[i], dims[i])


def test_gradient_shape(coupling):
    """Test the shape of the gradient."""
    # TODO: Test if the coupling is not optimized
    dim1 = np.random.randint(2, 7)
    dim2 = np.random.randint(2, 7)
    dims = [dim1, dim2]
    coup = coupling(dim1, dim2, is_longitudinal=False)
    grads = coup.get_coupling_gradients()
    for grad in grads:
        for term in grad:
            for i, ops in enumerate(term):
                assert np.size(ops) == 0
    coup.set_optimizable_parameters(coup.get_parameters())
    grads = coup.get_coupling_gradients()
    for grad in grads:
        for term in grad:
            for i, ops in enumerate(term):
                assert np.shape(ops) == (dims[i], dims[i])

    # Test with RWA
    dim1 = np.random.randint(2, 7)
    dim2 = np.random.randint(2, 7)
    dims = [dim1, dim2]
    coup = coupling(dim1, dim2, is_longitudinal=False, use_rwa=True)
    coup.set_optimizable_parameters(coup.get_parameters())
    grads = coup.get_coupling_gradients()
    for grad in grads:
        for term in grad:
            for i, ops in enumerate(term):
                assert np.shape(ops) == (dims[i], dims[i])



"""Data class definition for the optimization result object."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

from paraqeet.file_logger import Logger
from paraqeet.measurement.measurement import Measurement
from paraqeet.optimization_map import OptimizationMap
from paraqeet.quantity import Array


@dataclass(repr=False)
class OptimizationResult:
    """Data class for representing optimization results.

    Attributes
    ----------
    STATUS_FINISHED : int
        The optimization finished without a clear success or failure.
        This is used by algorithms that do not necessarily converge towards a
        solution.
    STATUS_SUCCESS : int
        The optimization successfully found an optimum.
    STATUS_FAILED : int
        The optimization failed to converge.
    status : int
        Indicates if the optimization was successful.
        Should have one of the status constants as value.
    value : float
        The value at the best point of the optimized function.
    iterations : int
        The number of iterations during the optimization.
    message : str | None, optional
        Any additional message from the optimization algorithm.
        This can be an error message in case of failure.
    raw_result : Any | None, optional
        The raw result from the underlying algorithm.

    """

    # The optimization finished without a clear success or failure.
    # This is used by algorithms that do not necessarily converge towards
    # a solution.
    STATUS_FINISHED = 0
    # The optimization successfully found an optimum.
    STATUS_SUCCESS = 1
    # The optimization failed to converge.
    STATUS_FAILED = 2

    # Indicates if the optimization was successful.
    # Should have one of the status constants as value.
    status: int
    # The value at the best point of the optimized function.
    value: float
    # The number of iterations during the optimization.
    iterations: int
    # Any additional message from the optimization algorithm.
    # This can be an error message in case of failure.
    message: str | None = None
    # The raw result from the underlying algorithm.
    raw_result: Any | None = None

    def __repr__(self):
        """Magic method for human-readable printable representation.

        Represents the Optimizer object as a dictionary with status,
        value, and iterations. If a message has been added, adds that
        too the dict too.

        """
        as_dict = {
            "status": self.status,
            "value": self.value,
            "iterations": self.iterations,
        }
        if self.message:
            as_dict["message"] = self.message

        return str(as_dict)


class Optimizer(ABC):
    """Base class for all classes that implement an optimization algorithm.

    The class accepts a list of optimizable parameters from the lower layers
    which shall be optimized in order to minimize the given measure.

    Parameters
    ----------
    measure: Measurement
        Implementation of the Measurement class that measures the observable
        to be minimized.
    optimizables: OptimizationMap
        An optimization map containing all parameters that can be optimized.
        If none, an empty map will be created to which the parameters can
        be added later used.
    logger: FileLogger
        The file logger object.

    """

    _measure: Measurement
    _optimization_map: OptimizationMap
    _opt_idxs: list[int]
    _logger: Logger | None
    _times: Array

    def __init__(
        self,
        measure: Measurement,
        optimization_map: OptimizationMap,
        logger: Logger | None = None,
    ):
        self._measure = measure
        self._logger = logger
        self.optimization_map = optimization_map

    @property
    def logger(self) -> Logger | None:
        """Returns the current logger that is being used by this optimizer, or None if no logger was set yet."""
        return self._logger

    @logger.setter
    def logger(self, logger: Logger):
        """Set the logger for the optimizer object.

        Parameters
        ----------
        logger : Logger
            Logger object to be set as the logger for the system.
        """
        self._logger = logger

    @property
    def optimization_map(self) -> OptimizationMap:
        """Return the optimization map that this optimizer uses.

        Parameters that can be optimized need to be added to this map.

        Returns
        -------
        paraqeet.optimization_map
            Returns the optimization map that this optimizer uses.

        """
        return self._optimization_map

    @optimization_map.setter
    def optimization_map(self, opt: OptimizationMap) -> None:
        """Set an optimization_map.

        Registers optimizables and their length to keep track of vector
        and matrix valued parameters.

        Parameters
        ----------
        opt: OptimizationMap
            Takes in the optimizables to set parameters.

        """
        self._optimization_map = opt

    @abstractmethod
    def optimize(self, times: Array | float) -> OptimizationResult:
        """Perform the actual optimization.

        Depending on the implementation, this function might take a long
        time and might need several calls to the underlying layers.
        The returned object contains some information about the result.
        The result will include the raw result of the underlying algorithm
        for more information.

        Returns
        -------
        OptimizationResult
            Result of optimization via the OptimizationResult object.
            (status, value, iterations and the raw result)

        """
        pass

    def _build_optimizable_index_list(self):
        """Build the optimizable index list.

        Register optimizables and their length to keep track of vector
        and matrix valued parameters.

        """
        params = self._optimization_map.get_all_parameters()
        self._opt_idxs = []
        index = 0
        for qty in params:
            index += qty.get_length()
            self._opt_idxs.append(index)



"""Test the identity propagation model."""

import jax.numpy as jnp

from paraqeet.propagation.differentiable_propagation import DifferentiablePropagation
from paraqeet.quantity import Array, Quantity


class IdentityPropagation(DifferentiablePropagation):
    """Mock identity propagation implementation.

    Returns the initial state as the target state.
    """

    _state: Array

    def __init__(self):
        super().__init__(None, None)

    def set_initial_state(self, state: Array):
        """Set the initial state of the system.

        Set it to the given state argument.

        Parameters
        ----------
        state : Array
            Given state to be set as the initial state.

        """
        self._state = state

    def propagate(self, time: Array) -> Array:
        """Get the propagated state across the timestamps.

        Parameters
        ----------
        time: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the propagated values of the state across timestamps.

        """
        return jnp.array([self._state] * len(time))

    def get_value_and_gradient(self, time: Array) -> tuple[Array, Array]:
        # Returns an empty gradient because the class has 0 parameters
        empty_gradient = jnp.zeros(shape=(len(time), 0, len(self._state)))
        return self.propagate(time), empty_gradient

    def get_parameters(self) -> list[Quantity]:
        """Returns an empty list."""
        return []



"""Class definition of the Dummy model for testing."""

from paraqeet.model.equation_of_motion import EquationOfMotion
from paraqeet.quantity import Array, Quantity


class DummyEquationsOfMotion(EquationOfMotion):
    """Dummy model class to construct derived model classes.

    Parameters
    ----------
    hamiltonian: Hamiltonian
        Class object for a matrix representation of a Hamiltonian.
    """

    def __init__(self, hamiltonian):
        super().__init__(hamiltonian)

    def get_parameters(self) -> list[Quantity]:
        """Get parameters of the system.

        Returns
        -------
        list[Quantity]
            List of parameters of the system.

        """
        return []

    def get_value(self, t: Array) -> Array:
        """Get the matrix representation of the equations of motion.

        Parameters
        ----------
        t: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the matrix equations of motion.

        """
        return -1.0j * self._hamiltonian.get_value(t)

    def get_value_and_gradient(self, times) -> Array:
        """Compute the gradient of get_value.

        Parameters
        ----------
        times: Array
            One-dimensional vector of timestamps.

        """
        return -1.0j * self._hamiltonian.get_gradient_at_timestep(times)



import numpy as np
import pytest

from paraqeet.measurement.smoothness import Smoothness
from paraqeet.optimization_map import OptimizationMap
from paraqeet.quantity import Quantity
from paraqeet.signal.envelopes import GaussEnvelope
from paraqeet.signal.pwc_generator import PWCGenerator


@pytest.fixture
def pwc_gen():
    delta_sampling = 33e-9
    n_pwc = 40  # number of piecewise constants in the pulse
    t_final = n_pwc * delta_sampling
    tlist = np.linspace(0, t_final, n_pwc + 1)
    eps_qubit = 2 * np.pi * 1.0  # (in MHz)
    eps_max_qubit = 5 * eps_qubit  # (in MHz)
    tone = GaussEnvelope(
        amplitude=Quantity(eps_qubit * 1e6, -eps_max_qubit * 1e6, eps_max_qubit * 1e6),
        t_final=Quantity(t_final, 1 / 2 * t_final, 2 * t_final),
    )
    return PWCGenerator(envelopes=[tone], max_amplitude=eps_max_qubit * 1e6, tlist=tlist)


@pytest.fixture
def another_tone():
    t_final = 10e-9
    eps = 2 * np.pi * 0.7  # (in MHz)
    eps_max = 5 * eps  # (in MHz)
    tone = GaussEnvelope(
        amplitude=Quantity(eps * 1e6, -eps_max * 1e6, eps_max * 1e6),
        t_final=Quantity(t_final, 1 / 2 * t_final, 2 * t_final),
    )
    return tone


def test_smoothness_init(pwc_gen):
    smoothness = Smoothness(pwc_generator=pwc_gen)
    assert smoothness is not None


def test_smoothness_measure(pwc_gen):
    smoothness = Smoothness(pwc_generator=pwc_gen)
    expected_measured_value = 0.9999547789963114
    assert np.abs(smoothness.calculate_normalized_scalar(pwc_gen._time_grid) - expected_measured_value) < 1e-8


def test_smoothness_gradient(pwc_gen, another_tone):
    smoothness = Smoothness(pwc_generator=pwc_gen)
    optmap = OptimizationMap()
    optmap.add(pwc_gen, pwc_gen.get_parameters())
    # We add dummy parameters to check if the gradient is computed
    # correctly by padding zeros
    # optmap.add(tone_qubit, tone_qubit.get_parameters())
    optmap.add(another_tone, another_tone.get_parameters())
    optmap.register_params_with_optimizables()
    assert len(optmap.get_all_parameters()) == len(another_tone.optimizable_parameters) + len(
        pwc_gen.optimizable_parameters
    )
    _, gradient = smoothness.get_value_and_gradient(times=pwc_gen.tlist)
    num_opt_params = 0
    for param in optmap.get_all_parameters():
        num_opt_params += param.get_value().shape[0]
    assert gradient.shape[0] == num_opt_params
    assert np.abs(gradient[6] - 5.39894375e-13) / 5.39894375e-13 < 1e-5  # precomputed value
    assert np.abs(gradient[39] - 4.626191950335811e-14) / 4.626191950335811e-14 < 1e-5  # precomputed value
    # since the gradient of smoothness does not depend
    # on the parameters of tone we can also check:
    assert gradient[-1] == 0.0
    assert gradient[-2] == 0.0



from importlib.metadata import version

from sphinx_pyproject import SphinxConfig

project = "ParaQeet"
__version__ = version("paraqeet")
release = __version__
config = SphinxConfig("../pyproject.toml", globalns=globals(), config_overrides={"version": __version__})

# Install package in editable mode so autodoc can import modules

extensions = [
    "sphinx.ext.napoleon",  # to parse numpy stye python docstrings
    "sphinx.ext.mathjax",  # to include math expressions in the .rst files
    "nbsphinx",  # to include jupyter notebooks,
    "IPython.sphinxext.ipython_console_highlighting",
    "sphinx.ext.autodoc",  # Core library for html generation from docstrings
    "sphinx.ext.autosummary",  # Create neat summary tables
    "myst_parser",  #  Include md in html
]

# Automatically extract typehints when specified and place them in
# descriptions of the relevant function/method.
autodoc_typehints = "description"

exclude_patterns = ["**.ipynb_checkpoints"]

html_theme = "pydata_sphinx_theme"

html_theme_options = {
    "logo": {
        "alt_text": "ParaQeet",
        "text": "ParaQeet",
        "image_light": "../logo.png",
        "image_dark": "../logo.png",
    },
    "gitlab_url": "https://jugit.fz-juelich.de/pgi-12-external/qfc/paraqeet",
    "show_nav_level": 1,
    "show_toc_level": 1,
    "show_prev_next": True,  # Enable prev/next buttons
    "collapse_navigation": True,
    "header_links_before_dropdown": 4,
    "navbar_end": ["search-button", "theme-switcher", "version-switcher", "navbar-icon-links"],
    "navbar_persistent": [],
    "show_version_warning_banner": True,
    "secondary_sidebar_items": ["page-toc"],  # show subheadings in sidebar
}


html_title = f"{project} v{release}"
htmlhelp_basename = "paraqeet"

# Disable “View page source” link for index page
html_show_sourcelink = False

# XeLaTeX for better support of unicode characters
latex_engine = "xelatex"

html_static_path = ["_static"]
html_css_files = ["custom.css"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]



"""Test the signal gradient functions."""

import jax.numpy as jnp
import numpy as np
import pytest

from paraqeet.quantity import Array
from paraqeet.signal.envelopes import FlatTopGaussianEnvelope
from tests.dummy_device import FlatTopGaussianEnvelopeAD

time = jnp.linspace(0, 10e-6, 100)


@pytest.fixture
def tone():
    """Tone with analytic gradient.

    Returns
    -------
    paraqeet.signal.envelopes.Envelope
        A Flat top Gaussian Envelope
    """
    return FlatTopGaussianEnvelope()


@pytest.fixture
def tone_ad():
    """Tone with AutoDiff gradients.

    Returns
    -------
    paraqeet.signal.envelopes.Envelope
        A Flat top Gaussian Envelope without gradients defined
    """
    return FlatTopGaussianEnvelopeAD()


def random_entries_from_list(elements: Array, num: int = 0) -> Array:
    """Get random entries from an array of elements.

    Parameters
    ----------
    elements: Array
        Array of elements to choose from.
    num: int
        Number of random entries asked for.

    Returns
    -------
    Array
        Returns random entries selected from a list of elements.

    """
    if num is None:
        num = np.random.randint(0, len(elements))
    indices = np.random.choice(len(elements), num, replace=False)
    return np.array(elements)[indices]


def test_values(tone, tone_ad):
    """Test values from the AD signal."""
    assert tone_ad.get_value(time) == pytest.approx(tone.get_value(time), rel=1e-6)


def test_gradients(tone, tone_ad):
    """Test generation of analytic and AD signal gradients one at a time."""
    grad_tone = []
    grad_tone_ad = []

    # Pick a random number of parameters from the tones
    num_params = np.random.randint(0, len(tone.get_parameters()))
    for t in time:
        tone.set_optimizable_parameters(random_entries_from_list(tone.get_parameters(), num_params))
        tone_ad.set_optimizable_parameters(random_entries_from_list(tone_ad.get_parameters(), num_params))

        grad_tone_ad.append(tone_ad.get_value_and_gradient(t)[1])
        grad_tone.append(tone.get_value_and_gradient(t)[1])

    assert jnp.array(grad_tone_ad) == pytest.approx(jnp.array(grad_tone), rel=1e-6)


def test_gradients_vectorized(tone, tone_ad):
    """Test vectorized generation of analytic and AD signal gradients."""
    num_params = np.random.randint(0, len(tone.get_parameters()))

    tone_ad.set_optimizable_parameters(random_entries_from_list(tone_ad.get_parameters(), num_params))
    grad_tone_ad = tone_ad.get_value_and_gradient(time)[1]
    tone.set_optimizable_parameters(random_entries_from_list(tone.get_parameters(), num_params))
    grad_tone = tone.get_value_and_gradient(time)[1]
    assert jnp.array(grad_tone_ad) == pytest.approx(jnp.array(grad_tone), rel=1e-6)


def test_gradients_shape(tone):
    """Test the signal gradient shape."""
    tone.set_optimizable_parameters(random_entries_from_list(tone.get_parameters()))
    grads = tone.get_value_and_gradient(time)[1]
    assert grads.shape[0] == time.shape[0]



"""Class definition for the Device model."""

from abc import abstractmethod
from collections.abc import Callable
from functools import partial
from typing import Any

import jax
import jax.numpy as jnp
from jax import jacfwd, jit, vmap
from jax.scipy.special import erf

from paraqeet.differentiable import Differentiable
from paraqeet.optimizable import Optimizable
from paraqeet.quantity import Array, Quantity

jax.config.update("jax_enable_x64", True)


class Waveform(Optimizable, Differentiable):
    """Classical electronics."""

    _partial_grads_function: Callable | None = None
    _gradient_function: Callable | None = None
    _grad_arg_nums: tuple[int, ...] = ()

    def _compute_gradient_function(
        self, signal_function: Callable, argnums: tuple[int, ...], vmap_axes: tuple[int | None, ...]
    ):
        """Return a compute gradient function from the signal function.

        Parameters
        ----------
        signal_function: Callable
            A function that generated signals.
        argnums : Tuple[int, ...]
            A tuple of ints containing a variable number of argument numbers.
        vmap_axes : Tuple[int, ...]
            A tuple of ints.

        """
        grads = jacfwd(signal_function, argnums=argnums)
        self._partial_grads_function = grads
        partial_grads = vmap(grads, vmap_axes)
        self._gradient_function = jit(partial_grads)

    def set_optimizable_parameters(self, params: list[Quantity]) -> None:
        """Set optimizable parameters for optimization.

        Parameters
        ----------
        params: list[Quantity]
            Input list of parameters to be set.

        """
        super().set_optimizable_parameters(params)

        self._grad_arg_nums = ()
        for i, param in enumerate(self.get_parameters()):
            if self._is_optimized(param):
                self._grad_arg_nums += (i,)

        # Recompute gradient function
        params = self.get_parameters()
        num_params = len(params)

        # vmap over time axis only, set everything else to None
        vmap_axes = (None,) * num_params
        vmap_axes += (0,)  # type: ignore

        if len(self._grad_arg_nums) > 0:
            self._compute_gradient_function(
                self._evaluate,
                argnums=self._grad_arg_nums,
                vmap_axes=vmap_axes,
            )
        else:
            self._gradient_function = None

    @abstractmethod
    def _evaluate(self, *args, **kwargs) -> Array:
        """Evaluate the output of the system.

        *Note- It is recommended to make this function a 'pure' JAX function supporting JIT.*
        *The arguments are supposed to be arranged as (parameters, t), i.e., time after parameters.*
        *The type of arguments should be jax.Array.*
        *The output has to be a scalar for a scalar time input to support AD.*
        *Use jax.squeeze() to remove extra dimensions.*
        """
        pass

    @abstractmethod
    def get_value(self, times: Array | float) -> Array:
        """Compute the output.

        Parameters
        ----------
        times: Array
            One-dimensional vector of timestamps or a single value.

        Returns
        -------
        Array
            Output of the computation.

        """
        pass

    def get_value_and_gradient(self, times: Array | float) -> tuple[Array, Array]:
        """Compute the gradient of the `_evaluate` method.

        Uses Automatic differentiation as a fallback.
        The `_evaluate` method should be a `pure` function (should take the
        optimizable parameters as function arguments and doesn't depend on
        global variables).
        Refer to https://jax.readthedocs.io/en/latest/notebooks/Common_Gotchas_in_JAX.html
        for functionally `pure` functions.
        To implement analytical gradients / other methods for gradient
        computation overwrite this method in the inherited class.

        Parameters
        ----------
        times: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the gradient array of the `_evaluate` method.

        """
        params = self.get_parameters()
        param_values = [param.get_value() for param in params]
        t_arr = jnp.array(times, ndmin=1)
        value = jnp.empty(t_arr.shape[0])
        grads = jnp.empty((t_arr.shape[0], 0))
        if self._gradient_function is not None:
            value = jit(self._evaluate)(*param_values, t_arr)
            grad = self._gradient_function(*param_values, t_arr)
            grads = jnp.stack(grad, axis=1)
            grads = jnp.squeeze(grads, -1)
        return value, grads

    def get_time_gradient(self, times: Array | float) -> Array:
        """Compute a signal envelopes time derivative.

        Parameters
        ----------
        times: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns a vector signals time derivative.

        """
        t_arr = jnp.array(times, ndmin=1)
        env_time_grad_fun = jacfwd(self.get_value, argnums=0)
        env_time_grad = vmap(env_time_grad_fun, in_axes=(0,))(t_arr)
        return jnp.squeeze(env_time_grad)

    def get_time_and_parameter_gradient(self, times: Array | float) -> Array:
        r"""Compute the double derivative with respect to parameter and time.

        This function computes $\\frac{\\partial^2 \\Omega}{\\partial t \\partial \alpha}$
        for a pulse $\\Omega(t)$ and parameter $\\alpha$.

        Parameters
        ----------
        times: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns a vector signals time derivative.

        """
        params = self.get_parameters()
        param_values = [param.get_value() for param in params]
        t_arr = jnp.array(times, ndmin=1)
        grads = jnp.empty((t_arr.shape[0], 0))
        if self._partial_grads_function is not None:
            env_time_and_param_grad_fun = jacfwd(self._partial_grads_function, argnums=-1)
            env_time_and_param_grad = env_time_and_param_grad_fun(*param_values, t_arr)
            grads = jnp.stack(env_time_and_param_grad, axis=1)
            grads = jnp.diagonal(grads, axis1=0, axis2=-1)
            grads = jnp.transpose(grads, axes=(2, 0, 1))
        return jnp.squeeze(grads, axis=-1)


class LocalOscillator(Waveform):
    """A local oscillators carrier signal.

    _lo_freq : Quantity
        The frequency of the carrier signal.

    """

    _lo_freq: Quantity

    def __init__(self, frequency: Quantity | None = None) -> None:
        self._lo_freq = frequency or Quantity(
            value=jnp.array(4.8e9 * 2 * jnp.pi),
            min_value=jnp.array(0.8 * 4.8e9 * 2 * jnp.pi),
            max_value=jnp.array(1.2 * 4.8e9 * 2 * jnp.pi),
            unit="Hz",
            name="lo_freq",
            two_pi=True,
        )

    def get_parameters(self) -> list[Quantity]:
        """Return device parameters.

        Returns
        -------
        list[Quantity]
            Returns the carrier signal frequency.
        """
        return [self._lo_freq]

    @property
    def frequency(self) -> Quantity:
        """Get The frequency of the constant oscillating tone.

        Returns
        -------
        Quantity
            The frequency of the tone.

        """
        return self._lo_freq

    @frequency.setter
    def frequency(self, frequency: Quantity) -> None:
        """Set The frequency of the constant oscillating tone.

        Parameters
        ----------
        freq: Quantity
            The frequency of the constant oscillating tone.

        """
        self._lo_freq = frequency

    @partial(jax.jit, static_argnums=(0,))
    def _evaluate(self, freq: Array, times: Array) -> Array:  # type: ignore
        """Calculate the unscaled carrier signal.

        Parameters
        ----------
        freq: Array
            The frequency of the carrier signal
        times: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            The unscaled the carrier signal.
        """
        return jnp.exp(1j * freq * times)

    def get_value(self, times: Array | float) -> Array:
        """Evaluate a carrier signal from an input time vector.

        Parameters
        ----------
        times: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns a vector carrier signal.
        """
        # returns JitWrapped
        return self._evaluate(self._lo_freq.get_value(), times)  # type: ignore

    def get_value_and_gradient(self, times: Array | float) -> tuple[Array, Array]:
        """Return the gradient wrt to frequency of carrier signal.

        Parameters
        ----------
        times: Array
            Array of time points to evaluate gradients at.

        Returns
        -------
        Array
            Gradient of tone wrt to frequency.
        """
        freq = self._lo_freq.get_value()
        t_arr = jnp.array(times, ndmin=1)

        grads = jnp.empty((t_arr.shape[0], 0))
        if self._is_optimized(self._lo_freq):
            grads = jnp.reshape(1j * t_arr * self._evaluate(freq, t_arr), (-1, 1))

        value = self._evaluate(self._lo_freq.get_value(), times)
        return value, grads

    def get_time_gradient(self, times: Array | float) -> Array:
        """Compute a signals time derivative.

        Parameters
        ----------
        times: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array or JitWrapped
            Returns a vector signals time derivative.

        """
        freq = self._lo_freq.get_value()
        # returns JitWrapped
        return 1j * times * self._evaluate(freq, times)  # type: ignore


class DRAGMixer(Waveform):
    """A DRAG mixed waveform signal.

    The DRAG component is calculated for a set of envelopes and added in
    orthogonal direction in the x-y plane.

    _envs: list[Envelope]
        The list of shape defining signal envelops.
    _deltas: list[Quantity]
        The delta parameter by which to shift the frequency of the DRAG
        component.
    """

    def __init__(
        self,
        envelopes: Waveform | list[Waveform],
        deltas: list[Quantity] | None = None,
    ) -> None:
        self._envs = envelopes if isinstance(envelopes, list) else [envelopes]
        DRAGMixer._add_deltas(self._envs, deltas)

    def get_parameters(self) -> list[Quantity]:
        """Return a list of parameters.

        Collects and returns a list of parameters from the tone, generator
        and the carrier signal.

        Returns
        -------
        list[Quantity]
            All Parameters describing the signal.
        """
        params = list()
        for tone in self._envs:
            params += tone.get_parameters()
            params += [DRAGMixer._get_tone_delta(tone)]
        return params

    def get_envelopes(self) -> list[Waveform]:
        """Return envelopes from the DRAGMixer."""
        return self._envs

    @staticmethod
    def _add_deltas(envelope_tones: list[Waveform], deltas: list[Quantity] | None) -> None:
        """Add a DRAG delta parameter Quantity to each envelope Tone.

        Parameters
        ----------
        envelope_tones: list[Waveform]
            The list of tones defining the total envelope.
        deltas : list[Quantity]
            A List of Quantities representing the delta parameters to add to
            each envelope Tone.

        Returns
        -------
        list[Waveform]
            The list of envelope Tones with the added delta parameters.
        """
        for ii, env_tone in enumerate(envelope_tones):
            env_tone.__setattr__(
                "_delta",
                deltas[ii]
                if deltas
                else Quantity(
                    jnp.array(-200e6 * 2 * jnp.pi),
                    min_value=jnp.array(-3 * 200e6 * 2 * jnp.pi),
                    max_value=jnp.array(-0.1 * 200e6 * 2 * jnp.pi),
                    unit="Hz",
                    name="Delta",
                ),
            )

    @staticmethod
    def _get_tone_delta(tone: Waveform) -> Any:
        """Return a list of deltas for each tone.

        Returns
        -------
        Quantity
            List of delta values for each tone.
        """
        return tone.__getattribute__("_delta")

    def _evaluate(self, times: Array | float, *deltas) -> Array:
        """Compute the DRAG Envelope using deltas.

        Explicit function depending on deltas to compute gradients using AD.

        Parameters
        ----------
        t: Array
            One-dimensional vector of timestamps.
        deltas: list[float]
            Variable number of inputs for delta parameters for each tone.

        Returns
        -------
        Array
            Returns a vector signal of the DRAG envelope.
        """
        total_env = jnp.zeros_like(times, dtype=jnp.complex128)
        for delta, tone in zip(deltas, self._envs):
            env = tone.get_value(times)
            env_grad = tone.get_time_gradient(times)
            total_env += env - 1.0j / delta * env_grad
        return jnp.squeeze(total_env)

    def get_value(self, times: Array | float) -> Array:
        """Evaluate a carrier signal from an input time vector.

        Parameters
        ----------
        t: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns a vector carrier signal.
        """
        deltas = [DRAGMixer._get_tone_delta(tone).get_value() for tone in self._envs]
        return self._evaluate(times, *deltas)

    def set_optimizable_parameters(self, params: list[Quantity]) -> None:
        """Set specified parameters to be optimized.

        Also add the indices to `_grad_arg_nums` to compute the gradients.

        Parameters
        ----------
        params: list[Quantity]
        """
        super().set_optimizable_parameters(params)

        for tone in self._envs:
            tone.set_optimizable_parameters(params)

    def get_value_and_gradient(self, times: Array | float) -> tuple[Array, Array]:
        """Generate gradient of the signal for an array of time.

        Collect and return the parameter gradients from the Tone and the carrier
        Tone. Compute the gradient of the generator parameters by AD.
        The order of the gradients should match the order of parameters in
        `self.get_parameter()` method

        Parameters
        ----------
        times: Array
            An array of time points.

        Returns
        -------
        Array
            Array of gradients wrt each parameter for each time point.
        """
        deltas = [DRAGMixer._get_tone_delta(tone) for tone in self._envs]
        delta_values = [delta.get_value() for delta in deltas]
        times_arr = jnp.array(times, ndmin=1)

        gradients = jnp.zeros(shape=(times_arr.shape[0], 0))

        # Collect gradients wrt envelope parameters
        for i, tone in enumerate(self._envs):
            _, grads = tone.get_value_and_gradient(times_arr)
            mixed_der = tone.get_time_and_parameter_gradient(times_arr)
            grads += -1.0j / delta_values[i] * mixed_der
            gradients = jnp.append(gradients, grads, axis=1)

        # Collect gradients wrt deltas
        for i, tone in enumerate(self._envs):
            if self._is_optimized(deltas[i]):
                grad = 1j / (delta_values[i] ** 2) * tone.get_time_gradient(times_arr)
                grad = jnp.expand_dims(grad, axis=1)
                gradients = jnp.append(gradients, grad, axis=1)

        return self.get_value(times_arr), jnp.array(gradients)


class FlatTopGaussianFilter(Waveform):
    """A shape filter that forces the pulse to smoothly start and end at zero.
    This filter multiplies the input pulse with a flat-top Gaussian pulse.

    *Note - Use filters before the generators. Else Automatic differentiation does not work.*

    This is similar to `PWCGenerator.multiply_flat_top = True`.
    """

    _envs: list[Waveform]
    _t_final: Quantity

    def __init__(self, envelopes: Waveform | list[Waveform], t_final: Quantity):
        self._envs = envelopes if isinstance(envelopes, list) else [envelopes]
        self._t_final = t_final

    def get_parameters(self) -> list[Quantity]:
        """Return a list of parameters.

        Collects and returns a list of parameters from the tone, generator
        and the carrier signal.

        Returns
        -------
        list[Quantity]
            All Parameters describing the signal.
        """
        params = list()
        for tone in self._envs:
            params += tone.get_parameters()
        return params

    def get_envelopes(self) -> list[Waveform]:
        """Return envelopes from the FlatTopGaussianFilter."""
        return self._envs

    def set_optimizable_parameters(self, params: list[Quantity]) -> None:
        """Set specified parameters to be optimized.

        Also add the indices to `_grad_arg_nums` to compute the gradients.

        Parameters
        ----------
        params: list[Quantity]
        """
        super().set_optimizable_parameters(params)

        for tone in self._envs:
            tone.set_optimizable_parameters(params)

    @partial(jit, static_argnums=(0,))
    def _compute_flat_top_envelope(self, t):
        t_final = self._t_final.get_value()
        ramp_time = t_final / 25
        ramp_up = 1 + erf((t - 2 * t_final / 20) / ramp_time)
        ramp_down = 1 + erf((-t + 18 * t_final / 20) / ramp_time)
        return ramp_up * ramp_down / 4

    def _evaluate(self, t):
        return self._compute_flat_top_envelope(t)

    def get_value(self, times: Array | float) -> Array:
        """Evaluate a carrier signal from an input time vector.

        Parameters
        ----------
        times: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns a vector carrier signal.
        """
        total_env: Array = jnp.zeros_like(times, dtype=jnp.complex128)
        for tone in self._envs:
            total_env += tone.get_value(times)

        flattop_env = self._compute_flat_top_envelope(times)
        total_env *= flattop_env
        return jnp.squeeze(total_env)

    def get_value_and_gradient(self, times: Array | float) -> tuple[Array, Array]:
        """Generate gradient of the signal for an array of time.

        Collect and return the parameter gradients from the Tone and the carrier
        Tone. Compute the gradient of the generator parameters by AD.
        The order of the gradients should match the order of parameters in
        `self.get_parameter()` method

        Parameters
        ----------
        times: Array
            An array of time points.

        Returns
        -------
        Array
            Array of gradients wrt each parameter for each time point.
        """
        t_arr = jnp.array(times, ndmin=1)
        gradients = jnp.zeros(shape=(t_arr.shape[0], 0))
        total_env: Array = jnp.zeros_like(t_arr, dtype=jnp.complex128)
        smoothing = self._compute_flat_top_envelope(times)

        # Collect gradients wrt envelope parameters
        for tone in self._envs:
            value, grads = tone.get_value_and_gradient(times)
            total_env += value
            smoothing = jnp.reshape(smoothing, smoothing.shape + (1,) * (grads.ndim - smoothing.ndim))
            gradients = jnp.reshape(gradients, gradients.shape + (1,) * (grads.ndim - gradients.ndim))
            gradients = jnp.append(gradients, grads * smoothing, axis=1)
        return jnp.squeeze(total_env * smoothing), jnp.array(gradients)



"""Testing the Optimizables."""

import numpy as np

from paraqeet.optimizable import Optimizable
from paraqeet.quantity import Quantity


class DummyOptimizable(Optimizable):
    """An optimizable implementation.

    Does nothing except providing some random parameters.

    Parameters
    ----------
    random_quantity: Quantity
        New randomly generated Quantity object.
    num_params: int
        Number of parameters.
    """

    def __init__(self, random_quantity, num_params: int):
        super().__init__()
        self._optimizable_parameters = [random_quantity(np.random.randint(1, 20)) for i in range(num_params)]

    def get_parameters(self) -> list[Quantity]:
        """Get optimizable parameters."""
        return self._optimizable_parameters



"""Class definition for the Envelopes."""

import time
from abc import abstractmethod
from collections.abc import Callable
from functools import partial

import jax
import jax.numpy as jnp
from jax import jit
from jax.scipy.special import erf

from paraqeet.quantity import Array, Quantity
from paraqeet.signal.waveform import Waveform

jax.config.update("jax_enable_x64", True)


class Envelope(Waveform):
    """Classical Signal Envelope class.

    _amplitude: Quantity
        The amplitude of the envelope.
    _t_final: Quantity
        The length in time of the envelope.
    _gradient_function: Callable | None
        The function to calculate the gradient with respect to a set of
        previously defined parameters.
    _grad_arg_nums: tuple[int, ...]
        The identifying indices of which parameters to calculate the gradient
        with respect to.

    """

    _amplitude: Quantity
    _t_final: Quantity

    def __init__(
        self,
        amplitude: Quantity | None = None,
        t_final: Quantity | None = None,
    ):
        self._amplitude = amplitude or Quantity(
            1.55e8,
            min_value=jnp.array(0.0),
            max_value=jnp.array(1e9),
            unit="Hz",
            name="Amplitude",
            two_pi=True,
        )

        self._t_final = t_final or Quantity(
            32e-9,
            min_value=jnp.array(0),
            max_value=jnp.array(100e-9),
            unit="s",
            name="t_final",
        )

        self._gradient_function: Callable | None = None
        self._grad_arg_nums: tuple[int, ...] = ()

    def get_parameters(self):
        """Get a list of parameters of the envelope.

        Returns
        -------
        List[Quantity]
            List of parameters of the envelope.

        """
        return [self._amplitude, self._t_final]

    @property
    def amplitude(self) -> Quantity:
        """Get the amplitude of the system.

        Returns
        -------
        Quantity
            Amplitude of the system.

        """
        return self._amplitude

    @amplitude.setter
    def amplitude(self, amplitude: Quantity) -> None:
        """Set the amplitude of the system.

        Parameters
        ----------
        Quantity
            Amplitude value of the system to be set.

        """
        self._amplitude = amplitude

    @property
    def t_final(self) -> Quantity:
        """Get the length of the tone.

        Returns
        -------
        Quantity
            Length in time of the tone.

        """
        return self._t_final

    @t_final.setter
    def t_final(self, t_final: Quantity) -> None:
        """Set the length of the tone.

        Parameters
        ----------
        Quantity
            Length in time of the tone to be set.

        """
        self._t_final = t_final

    @abstractmethod
    def _evaluate(self, *args, **kwargs):
        """Evaluate the output of the envelope.

        Abstract method.

        Raises
        ------
        NotImplementedError
            Subclasses derived from this class must implement this method.

        """
        raise NotImplementedError()

    @abstractmethod
    def get_value(self, times: Array | float) -> Array:
        """Compute the output.

        Parameters
        ----------
        times: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Output of the computation.

        Raises
        ------
        NotImplementedError
            Subclasses derived from this class must implement this method.

        """
        raise NotImplementedError()


class ConstantEnvelope(Envelope):
    """A constant envelope tone with a fixed length.

    _amplitude: Quantity
        The amplitude of the envelope.
    _t_final: Quantity
        The length in time of the envelope.
    _gradientFunction: Callable | None
        The function to calculate the gradient with respect to a set of
        previously defined parameters.
    _grad_arg_nums: tuple[int, ...]
        The identifying indices of which parameters to calculate the gradient
        with respect to.

    """

    @partial(jit, static_argnums=(0,))
    def _evaluate(
        self,
        amp: Array,
        t_final: Array,
        t: Array | float,
    ) -> Array:
        """Evaluate the envelope depending on all parameters.

        Abstract method.

        Raises
        ------
        NotImplementedError
            Subclasses derived from this class must implement this method.

        """
        return jnp.squeeze(jnp.where(t <= t_final, amp, 0.0))

    def get_value(self, times: Array | float) -> Array:
        """Compute the constant signal envelope at different times.

        Parameters
        ----------
        times: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Output of the computation.

        Raises
        ------
        NotImplementedError
            Subclasses derived from this class must implement this method.

        """
        amp = self.amplitude.get_value()
        t_final = self.t_final.get_value()
        return self._evaluate(amp, t_final, times)  # type: ignore

    def get_time_gradient(self, times: Array | float) -> Array:
        """Compute a signal envelopes time derivative.

        Parameters
        ----------
        times: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns a vector signals time derivative.
        """
        return jnp.zeros_like(times)


class ZeroEnvelope(ConstantEnvelope):
    """Shorthand implementation of a zero signal envelope.

    _amplitude: Quantity
        The amplitude of the envelope.
    _t_final: Quantity
        The length in time of the envelope.
    _gradientFunction: Callable | None
        The function to calculate the gradient with respect to a set of
        previously defined parameters.
    _grad_arg_nums: tuple[int, ...]
        The identifying indices of which parameters to calculate the gradient
        with respect to.

    """

    def __init__(self):
        super().__init__()
        self.amplitude.set_value(0.0)


class FlatTopGaussianEnvelope(Envelope):
    """A flat-top Gaussian envelope.

    _amplitude: Quantity
        The amplitude of the envelope.
    _t_up: Quantity
        The start time of constant section of the envelope.
    _t_down: Quantity
        The end time of constant section of the envelope.
    _ramp_time: Quantity
        The rate of ramp up and ramp down of the envelope.
    _t_final: Quantity
        The length in time of the envelope. Used only if any of `t_up`, `t_down`,
        and `ramp_time` are none.
    _gradient_function: Callable | None
        The function to calculate the gradient with respect to a set of
        previously defined parameters.
    _grad_arg_nums: tuple[int, ...]
        The identifying indices of which parameters to calculate the gradient
        with respect to.
    """

    _amplitude: Quantity
    _t_final: Quantity
    _t_up: Quantity
    _t_down: Quantity
    _ramp_time: Quantity

    def __init__(
        self,
        amplitude: Quantity | None = None,
        t_up: Quantity | None = None,
        t_down: Quantity | None = None,
        ramp_time: Quantity | None = None,
        t_final: Quantity | None = None,
    ):
        self._amplitude = amplitude or Quantity(
            1.55e8,
            min_value=jnp.array(0.0),
            max_value=jnp.array(1e9),
            unit="Hz",
            name="Amplitude",
            two_pi=True,
        )

        self._t_final = t_final or Quantity(
            32e-9,
            min_value=jnp.array(0),
            max_value=jnp.array(100e-9),
            unit="s",
            name="t_final",
        )

        self._t_up = t_up or Quantity(
            self._t_final.get_value() / 5,
            min_value=self._t_final.get_min_value(),
            max_value=self._t_final.get_max_value(),
            unit="s",
            name="t_up",
        )

        self._t_down = t_down or Quantity(
            4 * self._t_final.get_value() / 5,
            min_value=self._t_final.get_min_value(),
            max_value=self._t_final.get_max_value(),
            unit="s",
            name="t_down",
        )

        self._ramp_time = ramp_time or Quantity(
            self._t_final.get_value() / 10,
            min_value=self._t_final.get_min_value(),
            max_value=self._t_final.get_max_value(),
            unit="s",
            name="ramp_time",
        )

        self._gradient_function: Callable | None = None
        self._grad_arg_nums: tuple[int, ...] = ()

    def get_parameters(self):
        """Get all parameters of the system."""
        return [self._amplitude, self._t_up, self._t_down, self._ramp_time]

    @partial(jit, static_argnums=(0,))
    def _evaluate(self, amp: Array, t_up: Array, t_down: Array, ramp_time: Array, t: Array):
        """Compute the output of the device.

        Explicitly depends on the optimizable parameters.

        Parameters
        ----------
        amp: Quantity
            Cosine pulse amplitude.
        t_up: Quantity
            The start time of constant section of the envelope.
        t_down: Quantity
            The end time of constant section of the envelope.
        ramp_time: Quantity
            The rate of ramp up and ramp down of the envelope.
        t: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the output of the device that explicitly depends
            on the optimizable parameters.

        """
        ramp_up = 1 + erf((t - t_up) / ramp_time)
        ramp_down = 1 + erf((-t + t_down) / ramp_time)
        return amp * ramp_up * ramp_down / 4

    @staticmethod
    @jit
    def _dir_erf(x: Array):
        return 2 / jnp.sqrt(jnp.pi) * jnp.exp(-(x**2))

    @partial(jit, static_argnums=(0,))
    def _evaluate_time_grad(self, amp: Array, t_up: Array, t_down: Array, ramp_time: Array, t: Array):
        """Compute the output of the device.

        Explicitly depends on the optimizable parameters.

        Parameters
        ----------
        amp: Quantity
            Cosine pulse amplitude.
        t_up: Quantity
            The start time of constant section of the envelope.
        t_down: Quantity
            The end time of constant section of the envelope.
        ramp_time: Quantity
            The rate of ramp up and ramp down of the envelope.
        t: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the output of the device that explicitly depends
            on the optimizable parameters.

        """
        ramp_up = 1 + erf((t - t_up) / ramp_time)
        ramp_up_t_dir = FlatTopGaussianEnvelope._dir_erf((t - t_up) / ramp_time)
        ramp_up_t_dir /= ramp_time

        ramp_down = 1 + erf((-t + t_down) / ramp_time)
        ramp_down_t_dir = FlatTopGaussianEnvelope._dir_erf((-t + t_down) / ramp_time)
        ramp_down_t_dir *= -1 / ramp_time

        prod_dir = ramp_up * ramp_down_t_dir + ramp_up_t_dir * ramp_down

        return amp * prod_dir / 4

    @partial(jit, static_argnums=(0,))
    def _evaluate_t_up_grad(self, amp: Array, t_up: Array, t_down: Array, ramp_time: Array, t: Array):
        """Compute the output of the device.

        Explicitly depends on the optimizable parameters.

        Parameters
        ----------
        amp: Quantity
            Cosine pulse amplitude.
        t_up: Quantity
            The start time of constant section of the envelope.
        t_down: Quantity
            The end time of constant section of the envelope.
        ramp_time: Quantity
            The rate of ramp up and ramp down of the envelope.
        t: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the output of the device that explicitly depends
            on the optimizable parameters.

        """
        ramp_up_dir = FlatTopGaussianEnvelope._dir_erf((t - t_up) / ramp_time)
        ramp_up_dir /= -ramp_time
        ramp_down = 1 + erf((-t + t_down) / ramp_time)
        return amp * ramp_up_dir * ramp_down / 4

    @partial(jit, static_argnums=(0,))
    def _evaluate_t_down_grad(self, amp: Array, t_up: Array, t_down: Array, ramp_time: Array, t: Array):
        """Compute the output of the device.

        Explicitly depends on the optimizable parameters.

        Parameters
        ----------
        amp: Quantity
            Cosine pulse amplitude.
        t_up: Quantity
            The start time of constant section of the envelope.
        t_down: Quantity
            The end time of constant section of the envelope.
        ramp_time: Quantity
            The rate of ramp up and ramp down of the envelope.
        t: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the output of the device that explicitly depends
            on the optimizable parameters.

        """
        ramp_up = 1 + erf((t - t_up) / ramp_time)
        ramp_down_dir = FlatTopGaussianEnvelope._dir_erf((-t + t_down) / ramp_time)
        ramp_down_dir /= ramp_time
        return amp * ramp_up * ramp_down_dir / 4

    @partial(jit, static_argnums=(0,))
    def _evaluate_ramp_time_grad(self, amp: Array, t_up: Array, t_down: Array, ramp_time: Array, t: Array):
        """Compute the output of the device.

        Explicitly depends on the optimizable parameters.

        Parameters
        ----------
        amp: Quantity
            Cosine pulse amplitude.
        t_up: Quantity
            The start time of constant section of the envelope.
        t_down: Quantity
            The end time of constant section of the envelope.
        ramp_time: Quantity
            The rate of ramp up and ramp down of the envelope.
        t: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the output of the device that explicitly depends
            on the optimizable parameters.

        """
        ramp_up = 1 + erf((t - t_up) / ramp_time)
        ramp_up_dir = FlatTopGaussianEnvelope._dir_erf((t - t_up) / ramp_time)
        ramp_up_dir *= -(t - t_up) / (ramp_time**2)

        ramp_down = 1 + erf((-t + t_down) / ramp_time)
        ramp_down_dir = FlatTopGaussianEnvelope._dir_erf((-t + t_down) / ramp_time)
        ramp_down_dir *= -(-t + t_down) / (ramp_time**2)

        prod_dir = ramp_up * ramp_down_dir + ramp_up_dir * ramp_down

        return amp * prod_dir / 4

    def get_value(self, times: Array | float) -> Array:
        """Get the output of the device on time stamps.

        Parameters
        ----------
        times: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the output of the device.

        """
        amp = self._amplitude.get_value()
        t_up = self._t_up.get_value()
        t_down = self._t_down.get_value()
        ramp_time = self._ramp_time.get_value()
        # returns JitWrapped
        return self._evaluate(amp, t_up, t_down, ramp_time, times)  # type: ignore

    def get_value_and_gradient(self, times: Array | float) -> tuple[Array, Array]:
        """Return the gradient wrt dimensionless parameters.

        Parameters
        ----------
        times: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the gradient wrt dimensionless parameters.

        """
        amp = self._amplitude.get_value()
        t_up = self._t_up.get_value()
        t_down = self._t_down.get_value()
        ramp_time = self._ramp_time.get_value()
        t_arr = jnp.array(times, ndmin=1)

        grads = []
        if self._is_optimized(self.amplitude):
            grads.append(self._evaluate(jnp.array([1.0]), t_up, t_down, ramp_time, t_arr))
        if self._is_optimized(self._t_up):
            grads.append(self._evaluate_t_up_grad(amp, t_up, t_down, ramp_time, t_arr))
        if self._is_optimized(self._t_down):
            grads.append(self._evaluate_t_down_grad(amp, t_up, t_down, ramp_time, t_arr))
        if self._is_optimized(self._ramp_time):
            grads.append(self._evaluate_ramp_time_grad(amp, t_up, t_down, ramp_time, t_arr))

        gradient = jnp.stack(grads, axis=1) if len(grads) > 0 else jnp.empty((t_arr.shape[0], 0))
        return self._evaluate(amp, t_up, t_down, ramp_time, times), gradient

    def get_time_gradient(self, times: Array | float) -> Array:
        """Compute a signal envelopes time derivative.

        Parameters
        ----------
        times: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns a vector signals time derivative.
        """
        amp = self.amplitude.get_value()
        t_up = self._t_up.get_value()
        t_down = self._t_down.get_value()
        ramp_time = self._ramp_time.get_value()
        return jnp.array(self._evaluate_time_grad(amp, t_up, t_down, ramp_time, times))


class GaussEnvelope(Envelope):
    """Create a simple Gauss envelope.

    _amplitude: Quantity
        The amplitude of the envelope.
    _t_final: Quantity
        The length in time of the envelope.
    _gradient_function: Callable | None
        The function to calculate the gradient with respect to a set of
        previously defined parameters.
    _grad_arg_nums: tuple[int, ...]
        The identifying indices of which parameters to calculate the gradient
        with respect to.

    """

    @partial(jax.jit, static_argnums=(0,))
    def _evaluate(self, amp: Array, t_final: Array, t: Array) -> Array:  # type: ignore
        """Calculate the unscaled gaussian signal.

        Parameters
        ----------
        t_final: Array
            Duration of the signal to calculate the center of the gaussian from.
        t: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            The unscaled gaussian signal.
        """
        sigma = t_final / 8
        env = amp * jnp.exp(-(1 / 2) * (t - t_final / 2) ** 2 / sigma**2)
        return jnp.squeeze(env)

    @partial(jax.jit, static_argnums=(0,))
    def _evaluate_time_gradient(self, amp: Array, t_final: Array, t: Array) -> Array:
        """Calculate the unscaled gaussian signal.

        Parameters
        ----------
        t_final: Array
            Duration of the signal to calculate the center of the gaussian from.
        t: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            The unscaled gaussian signals time derivative.
        """
        sigma = t_final / 8
        time_grad = self._evaluate(amp, t_final, t) * -1.0 * (t - t_final / 2) / sigma**2
        # returns JitWrapped
        return time_grad  # type: ignore

    def get_value(self, times: Array | float) -> Array:
        """Compute a Gaussian signal.

        Parameters
        ----------
        times: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns a vector gaussian signal.
        """
        t_final = self.t_final.get_value()
        amp = self.amplitude.get_value()
        # returns JitWrapped
        return self._evaluate(amp, t_final, times)  # type: ignore

    def get_time_gradient(self, times: Array | float) -> Array:
        """Compute a Gaussian signals time derivative.

        Parameters
        ----------
        times: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns a vector gaussian signals time derivative.
        """
        t_final = self.t_final.get_value()
        amp = self.amplitude.get_value()
        env_time_deriv = self._evaluate_time_gradient(amp, t_final, times)
        # returns JitWrapped
        return env_time_deriv  # type: ignore


class DCRABEnvelope(Envelope):
    r"""Create a dCRAB pulse envelope.

    The dCRAB pulse is given as a sum of sinusoidal components as [Müller2022]

    .. math::
        f(t) = g(t)( 1 + \sum_{i=1}^{N_c / 2} c_{2i} \frac{\cos(\omega_{2i} t)}{\Lambda(t)}
        + \sum_{i = 1} ^ {N_c/2} c_{2i + 1} \frac{\sin(\omega_{2i + 1} t)}{\Lambda(t)} )

    Here we consider :math:`g(t) = \Lambda(t) = 1` for simplicity.
    Further, even components are for cosine and odd components are for sine.
    *Note - The function is designed to work well for even total number of components.
    For odd total number it may not work as expected.*

    [Müller2022] Müller et al. "One decade of quantum optimal control in the chopped random basis"

    _amplitude: Quantity
        The amplitude of the envelope.
    _t_final: Quantity
        The length in time of the envelope.
    _num_components: int
        Number of components added each iteration to the dCRAB basis. Defaults to 2. Advised to be an even number.
    _total_num_components: int
        Total number of components in the current dCRAB basis. This is the number of coefficients
        or the number of frequencies present. NOT the sum of them.
    _real_coefficients: list[Quantity]
        Vector quantity as a list of amplitudes of individual sinusoidal components.
    _real_frequencies: list[Quantity]
        Vector quantity as a list of frequencies of individual sinusoidal components.
    _real_phases: list[Quantity]
        Vector quantity as a list of phases of individual sinusoidal components.
    _gradient_function: Callable | None
        The function to calculate the gradient with respect to a set of
        previously defined parameters.
    _grad_arg_nums: tuple[int, ...]
        The identifying indices of which parameters to calculate the gradient
        with respect to.
    """

    _amplitude: Quantity
    _t_final: Quantity
    _num_components: int
    _total_num_components: int
    _real_coefficients: list[Quantity]
    _real_frequencies: list[Quantity]
    _real_phases: list[Quantity]
    _imag_coefficients: list[Quantity]
    _imag_frequencies: list[Quantity]
    _imag_phases: list[Quantity]
    _min_frequency: float
    _max_frequency: float

    def __init__(
        self,
        amplitude: Quantity | None = None,
        t_final: Quantity | None = None,
        num_components: int = 2,
        min_frequency: float = 0.0,
        max_frequency: float = 2 * jnp.pi * 5.0,
        seed: int | None = None,
    ):
        self._amplitude = amplitude or Quantity(
            1.55e8,
            min_value=jnp.array(0.0),
            max_value=jnp.array(1e9),
            unit="Hz",
            name="Amplitude",
            two_pi=True,
        )

        self._t_final = t_final or Quantity(
            32e-9,
            min_value=jnp.array(0),
            max_value=jnp.array(100e-9),
            unit="s",
            name="t_final",
        )
        self._min_frequency = min_frequency
        self._max_frequency = max_frequency

        self._num_components = num_components

        if seed is None:
            seed = int(1e7 * time.time())

        key = jax.random.key(seed)
        coeffs = jax.random.uniform(key, shape=(self._num_components,), minval=-1, maxval=1)

        self._real_coefficients = [
            Quantity(
                coeffs[i],
                min_value=jnp.array(-1.0),
                max_value=jnp.array(1.0),
                unit="",
                name=f"CRAB Re coefficient {i}",
            )
            for i in range(self._num_components)
        ]

        key = jax.random.key(2 * seed + 91)
        coeffs = jax.random.uniform(key, shape=(self._num_components,), minval=-1, maxval=1)

        self._imag_coefficients = [
            Quantity(
                coeffs[i],
                min_value=jnp.array(-1.0),
                max_value=jnp.array(1.0),
                unit="",
                name=f"CRAB Im coefficient {i}",
            )
            for i in range(self._num_components)
        ]

        key = jax.random.key(3 * seed + 81)
        freqs = jax.random.uniform(
            key, shape=(self._num_components,), minval=self._min_frequency, maxval=self._max_frequency
        )

        self._real_frequencies = [
            Quantity(
                freqs[i],
                min_value=jnp.array(self._min_frequency),
                max_value=jnp.array(self._max_frequency),
                unit="Hz",
                name=f"CRAB Re frequency {i}",
                two_pi=True,
            )
            for i in range(self._num_components)
        ]

        key = jax.random.key(4 * seed + 19)
        freqs = jax.random.uniform(
            key, shape=(self._num_components,), minval=self._min_frequency, maxval=self._max_frequency
        )

        self._imag_frequencies = [
            Quantity(
                freqs[i],
                min_value=jnp.array(self._min_frequency),
                max_value=jnp.array(self._max_frequency),
                unit="Hz",
                name=f"CRAB Im frequency {i}",
                two_pi=True,
            )
            for i in range(self._num_components)
        ]

        key = jax.random.key(5 * seed + 73)
        phases = jax.random.uniform(key, shape=(self._num_components,), minval=-jnp.pi, maxval=jnp.pi)

        self._real_phases = [
            Quantity(
                phases[i],
                min_value=jnp.array(-jnp.pi),
                max_value=jnp.array(jnp.pi),
                unit="Hz",
                name=f"CRAB Re Phase {i}",
                two_pi=True,
            )
            for i in range(self._num_components)
        ]

        key = jax.random.key(6 * seed + 74)
        phases = jax.random.uniform(key, shape=(self._num_components,), minval=-jnp.pi, maxval=jnp.pi)

        self._imag_phases = [
            Quantity(
                phases[i],
                min_value=jnp.array(-jnp.pi),
                max_value=jnp.array(jnp.pi),
                unit="Hz",
                name=f"CRAB Im Phase {i}",
                two_pi=True,
            )
            for i in range(self._num_components)
        ]

        self._total_num_components = self._num_components

        self._gradient_function: Callable | None = None
        self._grad_arg_nums: tuple[int, ...] = ()

    def get_parameters(self):
        """Return the parameters of the CRAB signal.

        The parameters are arranged as follows,
        `[amplitude, t_final, ... total_num coefficients ..., ... total_num frequencies ...]`

        """
        params = [self.amplitude, self._t_final]
        params.extend(self._real_coefficients)
        params.extend(self._real_frequencies)
        params.extend(self._real_phases)
        params.extend(self._imag_coefficients)
        params.extend(self._imag_frequencies)
        params.extend(self._imag_phases)
        return params

    def get_coefficients_frequencies_and_phases(self) -> list[Quantity]:
        """Return all the coefficients and frequencies used in the CRAB signal."""
        return (
            self._real_coefficients
            + self._real_frequencies
            + self._real_phases
            + self._imag_coefficients
            + self._imag_frequencies
            + self._imag_phases
        )

    def add_new_components(self, seed: int | None = None) -> None:
        """Add `self._num_components` number of new randomized components to the optimization."""
        if seed is None:
            seed = int(1e7 * time.time())

        # Limit max coeff value so that the new components do not derail the optimization.
        key = jax.random.key(seed)
        coeffs = jax.random.uniform(key, shape=(self._num_components,), minval=-0.5, maxval=0.5)

        self._real_coefficients.extend(
            [
                Quantity(
                    coeffs[i],
                    min_value=jnp.array(-1.0),
                    max_value=jnp.array(1.0),
                    unit="",
                    name=f"CRAB Re coefficient {i + self._total_num_components}",
                )
                for i in range(self._num_components)
            ]
        )

        key = jax.random.key(4 * seed + 59)
        freqs = jax.random.uniform(
            key, shape=(self._num_components,), minval=self._min_frequency, maxval=self._max_frequency
        )

        self._real_frequencies.extend(
            [
                Quantity(
                    freqs[i],
                    min_value=jnp.array(self._min_frequency),
                    max_value=jnp.array(self._max_frequency),
                    unit="Hz",
                    name=f"CRAB Re frequency {i + self._total_num_components}",
                    two_pi=True,
                )
                for i in range(self._num_components)
            ]
        )

        key = jax.random.key(8 * seed + 61)
        phases = jax.random.uniform(key, shape=(self._num_components,), minval=-jnp.pi, maxval=jnp.pi)

        self._real_phases.extend(
            [
                Quantity(
                    phases[i],
                    min_value=jnp.array(-jnp.pi),
                    max_value=jnp.array(jnp.pi),
                    unit="Hz",
                    name=f"CRAB Re Phase {i + self._total_num_components}",
                    two_pi=True,
                )
                for i in range(self._num_components)
            ]
        )

        key = jax.random.key(10 * seed)
        coeffs = jax.random.uniform(key, shape=(self._num_components,), minval=-0.5, maxval=0.5)

        self._imag_coefficients.extend(
            [
                Quantity(
                    coeffs[i],
                    min_value=jnp.array(-1.0),
                    max_value=jnp.array(1.0),
                    unit="",
                    name=f"CRAB Im coefficient {i + self._total_num_components}",
                )
                for i in range(self._num_components)
            ]
        )

        key = jax.random.key(7 * seed + 89)
        freqs = jax.random.uniform(
            key, shape=(self._num_components,), minval=self._min_frequency, maxval=self._max_frequency
        )

        self._imag_frequencies.extend(
            [
                Quantity(
                    freqs[i],
                    min_value=jnp.array(self._min_frequency),
                    max_value=jnp.array(self._max_frequency),
                    unit="Hz",
                    name=f"CRAB Im frequency {i + self._total_num_components}",
                    two_pi=True,
                )
                for i in range(self._num_components)
            ]
        )

        key = jax.random.key(21 * seed + 34)
        phases = jax.random.uniform(key, shape=(self._num_components,), minval=-jnp.pi, maxval=jnp.pi)

        self._imag_phases.extend(
            [
                Quantity(
                    phases[i],
                    min_value=jnp.array(-jnp.pi),
                    max_value=jnp.array(jnp.pi),
                    unit="Hz",
                    name=f"CRAB Im phase {i + self._total_num_components}",
                    two_pi=True,
                )
                for i in range(self._num_components)
            ]
        )

        self._total_num_components += self._num_components

    def remove_small_coefficients(self, tol: float = 1e-7) -> None:
        """Remove coefficients (and corresponding frequencies) that are smaller than a tolerance."""
        removed_coeffs = []
        for i, real_coeff, imag_coeff in zip(
            range(self._total_num_components), self._real_coefficients, self._imag_coefficients
        ):
            if jnp.abs(real_coeff.get_value() + 1j * imag_coeff.get_value()) < tol:
                removed_coeffs.append(i)

        self._real_coefficients = [coeff for i, coeff in enumerate(self._real_coefficients) if i not in removed_coeffs]
        self._real_frequencies = [freq for i, freq in enumerate(self._real_frequencies) if i not in removed_coeffs]
        self._real_phases = [phase for i, phase in enumerate(self._real_phases) if i not in removed_coeffs]

        self._imag_coefficients = [coeff for i, coeff in enumerate(self._imag_coefficients) if i not in removed_coeffs]
        self._imag_frequencies = [freq for i, freq in enumerate(self._imag_frequencies) if i not in removed_coeffs]
        self._imag_phases = [phase for i, phase in enumerate(self._imag_phases) if i not in removed_coeffs]

        self._total_num_components = len(self._real_coefficients)

    @partial(jax.jit, static_argnums=(0,))
    def _evaluate(self, *params: Array) -> Array:  # type: ignore
        """Compute the CRAB pulse.

        Here the params is arranged as follows,
            [amplitude, t_final, ... total_num coefficients ..., ... total_num frequencies ..., t]

        This evaluate function is written in this way to make it compatible with adding new
        components and freezing existing components required for dCRAB optimization.
        """
        amp = params[0]
        t_final = params[1]
        real_coeffs: list[Array] = params[2 : 2 + self._total_num_components]  # type: ignore
        real_freqs: list[Array] = params[2 + self._total_num_components : 2 + 2 * self._total_num_components]  # type: ignore
        real_phases: list[Array] = params[2 + 2 * self._total_num_components : 2 + 3 * self._total_num_components]  # type: ignore

        imag_coeffs: list[Array] = params[2 + 3 * self._total_num_components : 2 + 4 * self._total_num_components]  # type: ignore
        imag_freqs: list[Array] = params[2 + 4 * self._total_num_components : 2 + 5 * self._total_num_components]  # type: ignore
        imag_phases: list[Array] = params[2 + 5 * self._total_num_components : 2 + 6 * self._total_num_components]  # type: ignore

        t = params[-1]
        env_real = jnp.zeros_like(t)
        env_imag = jnp.zeros_like(t)
        for i in range(self._total_num_components):
            env_real += real_coeffs[i] * jnp.cos(real_freqs[i] * t / t_final + real_phases[i])
            env_imag += imag_coeffs[i] * jnp.cos(imag_freqs[i] * t / t_final + imag_phases[i])
        env_real /= 2 * jnp.sum(jnp.abs(jnp.array(real_coeffs)))
        env_imag /= 2 * jnp.sum(jnp.abs(jnp.array(imag_coeffs)))

        env = env_real + 1j * env_imag
        return jnp.squeeze(amp * env)

    def get_value(self, t: Array) -> Array:
        """Compute the CRAB signal.

        Parameters
        ----------
        t: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns a vector CRAB signal.
        """
        t_final = self.t_final.get_value()
        amp = self.amplitude.get_value()
        real_coeffs = [coeff.get_value() for coeff in self._real_coefficients]
        real_freqs = [freq.get_value() for freq in self._real_frequencies]
        real_phases = [phase.get_value() for phase in self._real_phases]

        imag_coeffs = [coeff.get_value() for coeff in self._imag_coefficients]
        imag_freqs = [freq.get_value() for freq in self._imag_frequencies]
        imag_phases = [phase.get_value() for phase in self._imag_phases]

        params = [amp, t_final] + real_coeffs + real_freqs + real_phases + imag_coeffs + imag_freqs + imag_phases
        return self._evaluate(*params, t)  # type: ignore



"""Class definition of a two level system (TLS) for testing."""

import jax.numpy as jnp

from paraqeet.model.differentiable_hamiltonian import DifferentiableHamiltonian
from paraqeet.quantity import Array


class SpinRWA(DifferentiableHamiltonian):
    """A Single Spin."""

    def __init__(self, drives=None):
        super().__init__(drives)
        self.sigma_p = jnp.array([[0j, 1], [0, 0]])
        self.dim = 2

    def get_value_at_timestep(self, timestep: float) -> Array:
        """Just sigma-X."""
        return self._drives[0].get_value_at_timestep(self.sigma_p, timestep)

    def get_value_and_gradient(self, times: Array) -> tuple[Array, Array]:
        """Gradient is just the drive matrix."""
        return self.get_value(times), self._drives[0].get_gradient(self.sigma_p, times)

    def get_gradient_at_timestep(self, time):
        return self._drives[0].get_gradient_at_timestep(self.sigma_p, time)

    # TODO: implement dimension-method from Hamiltonian
    def dimension(self) -> int:
        raise NotImplementedError()

    #  TODO: implement get_collapseops method from Hamiltonian
    def get_collapseops(self) -> list[tuple[Array, Array]]:
        raise NotImplementedError()

    #  TODO: implement get_parameters method from Optimizable
    def get_parameters(self):
        raise NotImplementedError()



"""Testing the configuration functions."""

import numpy as np
import pytest
from scipy.stats import unitary_group

from paraqeet.quantity import Quantity
from tests.model.dummy_model import DummyEquationsOfMotion
from tests.model.empty_hamiltonian import EmptyHamiltonian

LEN_SIG = 20


@pytest.fixture
def ts():
    """Return evenly spaced numbers according to the given signal length.

    The given signal length `LEN_SIG` is a module level global variable.

    """
    return np.linspace(0, 1e-9, LEN_SIG)


@pytest.fixture
def identity():
    """Return an identity array generation function."""

    def _method(dimension):
        return np.identity(dimension)

    return _method


@pytest.fixture
def model():
    """Return a dummy model generation function."""

    def _method(dimension):
        return DummyEquationsOfMotion(EmptyHamiltonian(dimension))

    return _method


@pytest.fixture
def random_state():
    """Return a random state generating method.

    Generates random normalized states for a given dimension.

    """

    def _method(dimension):
        state = np.random.random(dimension) + 1j * np.random.random(dimension)
        return state / np.sqrt(np.vdot(state, state))

    return _method


@pytest.fixture
def random_matrix():
    """Return a random matrix generating method.

    Generates random matrix for given dimensions n and m.
    The matrix is normalized to have trace 1.

    """

    def _method(n, m):
        state = np.random.random(size=(n, m))
        +1j * np.random.random(size=(n, m))
        return state / np.trace(state)

    return _method


@pytest.fixture
def random_unitary_matrix():
    """Return a random unitary matrix generating method.

    Generates random unitary matrices for a given dimension.

    """
    seed = 84

    def _method(dim):
        return unitary_group.rvs(dim, random_state=seed)

    return _method


@pytest.fixture
def random_basis_vectors():
    """Return a random basis vector generating method.

    Generates num_vec vectors, each with 0 everywhere except a 1 at a random index.
    All vectors will be orthogonal.

    """

    def _method(dim, num_vec):
        v = np.zeros((dim, num_vec))
        indices = np.random.choice(np.arange(0, dim), num_vec, replace=False)
        for i in range(num_vec):
            v[indices[i], i] = 1
        return v

    return _method


@pytest.fixture
# helper functions
def random_quantity(random_quantity_for_values):
    """Generate a quantity with n positive and negative numbers.

    Each quantity is with the same order of magnitude
    which is chosen randomly between 1e-10 and 1e10.

    """

    def _method(n: int, unit: str = ""):
        magnitude = np.power(10.0, np.random.randint(-10, 10))
        values = (2 * np.random.random(n) - 1) * magnitude
        return random_quantity_for_values(values, unit)

    return _method


@pytest.fixture
def random_quantity_for_values(random_limits_for_quantity):
    """Generate a quantity from the given array of values.

    Generates while making sure that the limits are set correctly.

    """

    def _method(values: np.ndarray, unit: str = ""):
        limits = random_limits_for_quantity(values)
        return Quantity(values, min_value=limits[0], max_value=limits[1], unit=unit)

    return _method


@pytest.fixture
def random_limits_for_quantity():
    """Return random limits for quantities.

    Returns random but valid minimum and maximum values
    for the given value array while taking account for negative values.

    """

    def _method(values: np.ndarray):
        if len(values.shape) == 0:
            # scalar quantity
            if values == 0.0:
                min_value = np.array(-1.0)
                max_value = np.array(+1.0)
            elif values < 0:
                min_value = (np.random.random() + 1.1) * values
                max_value = np.random.random() * values
            else:
                min_value = np.random.random() * values
                max_value = (np.random.random() + 1.1) * values
            return min_value, max_value
        else:
            # list quantity
            min_values, max_values = (
                np.zeros_like(values),
                np.zeros_like(values),
            )
            for i, v in enumerate(values):
                if v == 0.0:
                    min_values[i] = -1
                    max_values[i] = +1
                elif v < 0:
                    min_values[i] = (np.random.random() + 1.1) * v
                    max_values[i] = np.random.random() * v
                else:
                    min_values[i] = np.random.random() * v
                    max_values[i] = (np.random.random() + 1.1) * v
            return min_values, max_values

    return _method


@pytest.fixture
def random_from_list():
    """Returns one or more random values from the given list, excluding specific values from the list."""

    def _method(select_from: np.ndarray, excluded: np.ndarray | float, num_selected: int = 1):
        if np.isscalar(excluded):
            excluded = np.array([excluded])
        selectable = np.delete(select_from, excluded)
        return np.random.choice(selectable)

    return _method



"""Class definition of the optimizable model."""

from abc import abstractmethod

from paraqeet.model.differentiable_hamiltonian import DifferentiableHamiltonian
from paraqeet.optimizable import Optimizable
from paraqeet.quantity import Array


class EquationOfMotion(Optimizable):
    """Represents the equation of motion for a given Hamiltonian.

    Implementations can for example be the Schrödinger equation for a
    closed system, Lindbladian for an open system, or Hamilton's equations
    for a classical system.

    Parameters
    ----------
    hamiltonian : Hamiltonian
        Matrix representation of a Hamiltonian.

    """

    _hamiltonian: DifferentiableHamiltonian
    # TODO: this is the first "public" variable used without the property-decorator.
    # Should we add a property for it?
    ode_propagation: bool = False

    def __init__(self, hamiltonian: DifferentiableHamiltonian):
        self._hamiltonian = hamiltonian

    def get_right_hand_side(self, time: Array, state: Array) -> Array:
        """Return the right-hand side of the equations of motion.

        The format depends on the implementation and could for example
        be a state vector or a matrix. Default implementation assumes a
        homogeneous ODE with matrix operator given by self.getMatrix().

        Parameters
        ----------
        time: Array
            Any one-dimensional vector of timestamps.

        Returns
        -------
        Array
            The right-hand side of the equation of motion at each time stamp.

        """
        return self.get_value(time) @ state

    @abstractmethod
    def get_value(self, times: Array) -> Array:
        """Abstract method to get the prefactor matrix.

        Parameters
        ----------
        times: Array
            Any one-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns matrix equations of motion.

        Raises
        ------
        NotImplementedError
            Subclasses derived from this class must implement this method.

        """
        pass

    # TODO: Since this method delegates the call to the Hamiltonian-instance, should
    # we rename it to get_habiltonian_value_and_gradient or similar? Otherwise it suggests
    # that it returns the gradient of the equation of motion itself.
    @abstractmethod
    def get_value_and_gradient(self, times: Array) -> tuple[Array, Array]:
        """Implement the gradient of either getEquationOfMotion or getMatrixEOM.

        Parameters
        ----------
        times: Array
            Any one-dimensional vector of timestamps.

        Raises
        ------
        NotImplementedError
            Subclasses derived from this class must implement this method.

        """
        pass



"""Custom Hamiltonian wrapper for H(t)."""

from collections.abc import Callable
from typing import Any

import jax.numpy as jnp
from jax import vmap

from paraqeet.exceptions import ConfigurationException
from paraqeet.model.differentiable_hamiltonian import DifferentiableHamiltonian
from paraqeet.quantity import Array, Quantity


class CustomHamiltonian(DifferentiableHamiltonian):
    """Custom Hamiltonian class to simulate systems using a user defined Hamitonian function..

    Here we expect a Hamiltonian function of the form `H(t, *params)`.
    Here `params` is a list of scalars (**NOT `Quantity`**).

    But, `parameters` is a list of `Quantity` that would be optimized.

    It is advised to make the Hamiltonian function vmap and jit compatible.
    Furthermore, it is advised to write the Hamiltonian function in a way such that
    it takes a single time point (scalar) as input and returns  a jax array of dimensions [n, n].

    Additionally, to optimize the parameters, one needs to pass a list of
    gradient functions corresponding to each parameter, in the same order as the parameter list.

    To use open system simulation, provide a list of tuples of decay rates and corresponding collapse operators.
    """

    _hamiltonian_function: Callable[[Array, Any], Array] | Callable[[float, Any], Array]
    _parameters: list[Quantity]
    _gradient_functions: list[Callable] | None
    _collapse_operatorss: list[tuple[Array, Array]] | None

    def __init__(
        self,
        hamiltonian_function: Callable[[Array, Any], Array] | Callable[[float, Any], Array],
        parameters: list[Quantity],
        gradient_functions: list[Callable] | None = None,
        collapse_operators: list[tuple[Array, Array]] | None = None,
    ):
        self._hamiltonian_function = hamiltonian_function
        self._parameters = parameters
        # TODO: is a leading underscore missing here?
        self.gradient_functions = gradient_functions
        self.collapse_operators = collapse_operators

    @property
    def gradient_functions(self) -> list[Callable] | None:
        """Return gradient functions."""
        return self._gradient_functions

    @gradient_functions.setter
    def gradient_functions(self, grad_funcs: list[Callable] | None):
        """Set gradient functions."""
        self._gradient_functions = grad_funcs

    @property
    def collapse_operators(self) -> list[tuple[Array, Array]] | None:
        """Return collapse operators."""
        return self._collapse_operators

    @collapse_operators.setter
    def collapse_operators(self, col_ops: list[tuple[Array, Array]] | None):
        """Set collapse operators."""
        self._collapse_operators = col_ops

    def dimension(self):
        """Return dimension of the Hilbert space."""
        return self.get_value_at_timestep(jnp.array([0.0])).shape[1]

    def get_parameters(self) -> list[Quantity]:
        """Return a list of optimizable parameters."""
        return self._parameters

    def get_value_at_timestep(self, timestep: float) -> Array:
        """Return Hamiltonian as a function of time for a single time point."""
        params = [p.get_value()[0] for p in self._parameters]
        return self._hamiltonian_function(timestep, *params)

    def get_value(self, times: Array) -> Array:
        """Return Hamiltonian as a function of time for an array of time."""
        params = [p.get_value()[0] for p in self._parameters]
        matrix_fun = vmap(self._hamiltonian_function, in_axes=(0,) + (None,) * len(params))
        # ignoring mypy due to vmap
        return matrix_fun(times, *params)  # type: ignore

    def get_gradient_at_timestep(self, timestep: float) -> Array:
        """Return the gradient as a function of time for a single time point."""
        params = [p.get_value()[0] for p in self._parameters]
        if self.gradient_functions is None:
            raise ConfigurationException("Specify the gradient functions of the Hamiltonian to compute gradients.")
        if len(self.gradient_functions) != len(params):
            raise ConfigurationException(
                f"Got {len(params)} parameters but got {len(self.gradient_functions)}. "
                + "Provide gradient methods for all the input parameters"
            )
        grads = jnp.array([grad_func(timestep, *params) for grad_func in self.gradient_functions])
        return grads

    def value_and_get_gradient_at_timestep(self, time) -> tuple[Array, Array] | tuple[float, Array]:
        """Return Hamiltonian and its gradient for one timestep."""
        return self.get_value_at_timestep(time), self.get_gradient_at_timestep(time)

    def get_value_and_gradient(self, times: Array) -> tuple[Array, Array]:
        """Return Hamiltonian and its gradient as a function of time."""
        # ignoring mypy due to vmap
        return self.get_value(times), jnp.array(vmap(self.get_gradient_at_timestep, in_axes=(0,))(times))  # type: ignore

    def get_collapseops(self) -> list[tuple[Array, Array]]:
        """Return collapse operators."""
        if self.collapse_operators is None:
            raise ConfigurationException("Collapse operators not specified.")
        return self.collapse_operators



"""Test the Euler propagation model."""

import numpy as np
import pytest

from paraqeet.propagation.euler import Euler
from tests.model.dummy_model import DummyEquationsOfMotion
from tests.model.empty_hamiltonian import EmptyHamiltonian
from tests.propagation.test_common_propagation import needs_initial_state


@pytest.fixture
def euler():
    """Return a Euler propagation model generating method."""

    def _method(dimension):
        return Euler(DummyEquationsOfMotion(EmptyHamiltonian(dimension)))

    return _method


def test_parameters(euler):
    """Test parameters of the model."""
    assert euler(2).get_parameters() == []


def test_state_dimension_vector(random_state, euler, ts):
    """Test the dimension and the norm of the state vector.

    The dimension and the norm should be the same.

    """
    for i in range(10):
        dim = np.random.randint(2, 30)
        state = random_state(dim)
        propagation = euler(dim)
        propagation.set_initial_state(state)
        propagatedStates = propagation.propagate(ts)
        assert propagatedStates.shape[0] == len(ts)
        assert propagatedStates.shape[1:] == state.shape


def test_state_dimension_matrix(random_matrix, euler, ts):
    """Test the state dimension matrix."""
    for _ in range(10):
        dim = np.random.randint(2, 30)
        state = random_matrix(dim, dim)
        propagation = euler(dim)
        propagation.set_initial_state(state)
        propagatedStates = propagation.propagate(ts)
        assert propagatedStates.shape[0] == len(ts)
        assert propagatedStates.shape[1:] == state.shape


def test_needs_initial_state(random_state, euler):
    for dim in range(2, 10):
        propagation = euler(dim)
        needs_initial_state(propagation, dim)



"""Test the unitary fidelity model."""

import numpy as np
import pytest

from paraqeet.measurement.unitary_fidelity import UnitaryFidelity
from tests.propagation.identity_propagation import IdentityPropagation
from tests.propagation.random_propagation import RandomPropagation


@pytest.fixture
def identity_propagation():
    """Return a mock identity propagation object."""
    return IdentityPropagation()


def test_positivity(random_unitary_matrix, random_basis_vectors):
    """Test that the fidelity is always positive."""
    for dim in range(5, 10):
        for i in range(50):
            propagation = RandomPropagation(dim, True)
            gate = random_unitary_matrix(dim)
            measurement = UnitaryFidelity(propagation, gate)
            times = np.array([1.0])
            m = measurement.measure(times=times)
            assert 0.0 <= m

            # gate must have at least two dimensions
            subDim = np.random.randint(2, dim)
            basisStates = random_basis_vectors(dim, subDim)
            # gate is defined on the subspace.
            gate = random_unitary_matrix(subDim)
            measurement = UnitaryFidelity(propagation, gate, basisStates)
            m = measurement.measure(times=times)
            assert 0.0 <= m


def test_equality(identity_propagation, random_unitary_matrix):
    """Test that F(U,U) = 1."""
    for dim in range(2, 10):
        for i in range(50):
            gate = random_unitary_matrix(dim)
            np.testing.assert_almost_equal(np.conjugate(gate.T) @ gate, np.eye(dim))
            propagation = IdentityPropagation()
            propagation.set_initial_state(gate)
            measurement = UnitaryFidelity(propagation, gate)
            m = measurement.measure(times=np.array([1.0]))
            np.testing.assert_almost_equal(m, 1.0)


def test_projection(identity_propagation, random_basis_vectors):
    """Test the projection of the state vectors."""
    for dim in range(2, 10):
        for i in range(50):
            gate = np.eye(dim)
            init_state = random_basis_vectors(dim + 4, dim)
            propagation = identity_propagation
            propagation.set_initial_state(gate)
            measurement = UnitaryFidelity(propagation, gate, basis_states=init_state)
            m = measurement.measure(times=np.array([1.0]))
            np.testing.assert_almost_equal(m, 1.0)


def test_incompatible_shape(identity_propagation, random_unitary_matrix):
    """Test incompatible shapes for initial and target states.

    A set of initial and target states with different dimensions
    should raise an exception.

    Raises
    ------
    Exception
        Different dimensions for the initial and target states
        raise an exception.

    """
    allDims = np.arange(2, 10)
    for dim in allDims:
        for i in range(50):
            gate = random_unitary_matrix(dim)

            # create a propagator of a different dimension
            dimensions = np.delete(allDims, np.where(allDims == dim)[0][0])
            propagation = RandomPropagation(np.random.choice(dimensions), True)

            measurement = UnitaryFidelity(propagation, gate)
            with pytest.raises(Exception):
                measurement.measure(times=np.array([1.0]))



import jax.numpy as jnp
import numpy.random as random
import pytest

from paraqeet.measurement.state_transfer_fidelity import StateTransferFidelity
from paraqeet.model.closed_system import ClosedSystem
from paraqeet.model.custom_hamiltonian import CustomHamiltonian
from paraqeet.optimization_map import OptimizationMap
from paraqeet.optimizers.scipy_optimizer_gradient import ScipyOptimizerGradient
from paraqeet.propagation.scipy_expm_goat import ScipyExpmGOAT
from paraqeet.quantity import Array, Quantity

sigma_x = jnp.array([[0j, 1], [1, 0]])
sigma_z = jnp.diag(jnp.array([1.0, -1.0]))
FREQ = 4.8e9 * 2 * jnp.pi
T_FINAL = 10e-9


amplitude = Quantity(
    value=jnp.array(1.55e8),
    min_value=jnp.array(0.0),
    max_value=jnp.array(1e9),
    unit="Hz",
    name="Amplitude",
    two_pi=True,
)

frequency = Quantity(
    value=jnp.array(4.8e9 * 2 * jnp.pi),
    min_value=jnp.array(0.8 * 4.8e9 * 2 * jnp.pi),
    max_value=jnp.array(1.2 * 4.8e9 * 2 * jnp.pi),
    unit="Hz",
    name="lo_freq",
    two_pi=True,
)


@pytest.fixture
def cos_envelope():
    """Define a cosine envelope."""

    def _method(t, amp, freq):
        return amp * jnp.cos(freq * t)

    return _method


@pytest.fixture
def tls_hamiltonian(cos_envelope):
    """Define a Two level system Hamiltonian."""

    def _method(t, amp, freq):
        return 0.5 * FREQ * sigma_z + sigma_x * cos_envelope(t, amp, freq)

    return _method


@pytest.fixture
def gradient_functions():
    """Define the gradient of Hamiltonian wrt amp and freq."""

    def grad_amp(t, amp, freq):
        """Gradient of Hamiltonian wrt amplitude."""
        return sigma_x * jnp.cos(freq * t)

    def grad_frequency(t, amp, freq):
        """Gradient of Hamiltonian wrt frequency."""
        return -sigma_x * amp * t * jnp.sin(freq * t)

    analytical_grad_funcs = [grad_amp, grad_frequency]
    return analytical_grad_funcs


@pytest.fixture
def tls(tls_hamiltonian, gradient_functions):
    """Closed system model."""
    tls_hamil = CustomHamiltonian(
        hamiltonian_function=tls_hamiltonian, parameters=[amplitude, frequency], gradient_functions=gradient_functions
    )
    return tls_hamil


@pytest.fixture
def fid(tls):
    """Get the fidelity function"""
    model = ClosedSystem(tls)
    prop = ScipyExpmGOAT(model, resolution=100e9)

    init = jnp.array([[1.0], [0]])  # |0>
    target = jnp.array([[0.0], [1]])  # |1>
    zeroone = StateTransferFidelity(propagation=prop, initial_state=init, target_state=target)
    return zeroone


@pytest.fixture
def opt(tls, fid):
    optmap = OptimizationMap()
    optmap.add(tls)
    opt = ScipyOptimizerGradient(fid, optimization_map=optmap)
    return opt


def test_optimization(opt):
    """Test gradient based optimization."""
    res = opt.optimize(times=T_FINAL)
    assert res.value < 1e-4


def test_shapes(random_matrix):
    for dim in range(2, 10):

        def generator(time) -> Array:
            mat = random_matrix(dim, dim)
            return mat + jnp.conjugate(mat).T

        hamil = CustomHamiltonian(hamiltonian_function=generator, parameters=[])
        assert hamil.dimension() == dim

        times = jnp.linspace(0, random.randint(1, 100) * random.random(), random.randint(2, 20))
        assert hamil.get_value_at_timestep(times[-1]).shape == (dim, dim)
        assert hamil.get_value(times).shape == (len(times), dim, dim)



"""Class definition for the Sinusoidal generator model."""

import jax.numpy as jnp

from paraqeet.quantity import Array, Quantity
from paraqeet.signal.generator import Generator
from paraqeet.signal.waveform import LocalOscillator, Waveform


class IQMixer(Generator):
    """Control signal generation.

    Waveforms of envelopes (low bandwidth) are mixed with a local oscillator
    (high bandwidth) to apply a desired control field to the system.

    Parameters
    ----------
    envelopes : List[Waveform]
        List of input devices.
    frequency : Quantity | None
        Frequency of local oscillator.
    phase : Quantity | None
        Phase of local oscillator.
    """

    _envs: list[Waveform]
    _phase: Quantity
    _optimizable_parameters: list[Quantity] = []

    def __init__(
        self,
        envelopes: list[Waveform] | None,
        frequency: Quantity | None = None,
        phase: Quantity | None = None,
    ):
        self._envs = envelopes or []

        self._lo = LocalOscillator(frequency=frequency)

        self._phase = phase or Quantity(
            jnp.array(0.0),
            min_value=jnp.array(-jnp.pi),
            max_value=jnp.array(jnp.pi),
            unit="rad",
            name="Phase",
        )

    def get_parameters(self) -> list[Quantity]:
        """Return a list of parameters.

        Collects and returns a list of parameters from the tone, generator
        and the carrier signal.

        Returns
        -------
        list[Quantity]
            All Parameters describing the signal.
        """
        pars = []
        for env in self._envs:
            pars += env.get_parameters()
        pars += self._lo.get_parameters()
        pars += [self._phase]
        return pars

    def set_optimizable_parameters(self, params: list[Quantity]) -> None:
        """Set specified parameters to be optimized.

        Parameters
        ----------
        params : list[Quantity]
        """
        super().set_optimizable_parameters(params)

        for dev in self._envs:
            dev.set_optimizable_parameters(params)

        self._lo.set_optimizable_parameters(params)

    def _complex_signal(self, times: Array | float) -> Array:
        """Generate a signal for time(s).

        Doesn't take real value now for ease of gradient computation.

        Parameters
        ----------
        times: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the signal vector.

        """
        env = jnp.zeros_like(times)
        for dev in self._envs:
            env += jnp.reshape(dev.get_value(times), env.shape)
        sig = env.conj() * self._lo.get_value(times)
        sig = sig * jnp.exp(-1j * self._phase.get_value())
        return sig

    def get_value(self, times: Array | float) -> Array:
        """Generate a signal for time(s).

        Parameters
        ----------
        t: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the signal vector.

        """
        return jnp.real(self._complex_signal(times))

    def get_value_and_gradient(self, times) -> tuple[Array, Array]:
        r"""Collect and returns the gradients from all devices.

        Since the

        .. math::
            signal = \Re(\epsilon(t)^*  \exp(i \omega t)  \exp(-i \phi))

        Derivative of the signal wrt optimizable parameter of envelope would be

        .. math::
            0.5 * \Re(\partial \epsilon(t)^* \exp(i \omega t)  \exp(-i \phi))

        (TODO - Check the envelope derivatives)

        And derivative of signal wrt parameter of LO would be

        .. math::
            0.5 i t \epsilon(t)^* \exp(i \omega t)  \exp(-i \phi))

        And derivative of signal wrt phase would be

        .. math::
            -0.5 i \epsilon(t)^* \exp(i \omega t)  \exp(-i \phi))

        The 0.5 are due to the Wirtinger derivatives due to Re part.

        Parameters
        ----------
        t: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the signal gradient vector.

        """
        phase_fac = jnp.exp(-1j * self._phase.get_value())
        lo_out = self._lo.get_value(times)
        sig = self._complex_signal(times)
        gradients = jnp.zeros(shape=(times.shape[0], 0))

        # Collect gradients for envelopes
        for dev in self._envs:
            _, grad = dev.get_value_and_gradient(times)
            grad = grad.conj()
            if grad.size != 0:
                grad *= jnp.expand_dims(lo_out * phase_fac, axis=1)
            gradients = jnp.append(gradients, 0.5 * jnp.real(grad), axis=1)

        # Collect LO gradients
        lo_freq = self._lo.get_parameters()[0]
        if self._is_optimized(lo_freq):
            gradients = jnp.append(
                gradients,
                jnp.expand_dims(0.5j * times * sig, 1),
                axis=1,
            )

        # Collect gradient of Phase
        if self._is_optimized(self._phase):
            gradients = jnp.append(
                gradients,
                jnp.expand_dims(-0.5j * sig, 1),
                axis=1,
            )
        return jnp.real(sig), gradients

    def get_gradient_at_timestep(self, time: float) -> Array:
        r"""Return the gradients from all devices at the given time.

        Since the

        .. math::
            signal = \Re(\epsilon(t)^*  \exp(i \omega t)  \exp(-i \phi))

        Derivative of the signal wrt optimizable parameter of envelope would be

        .. math::
            0.5 * \Re(\partial \epsilon(t)^* \exp(i \omega t)  \exp(-i \phi))

        (TODO - Check the envelope derivatives)

        And derivative of signal wrt parameter of LO would be

        .. math::
            0.5 i t \epsilon(t)^* \exp(i \omega t)  \exp(-i \phi))

        And derivative of signal wrt phase would be

        .. math::
            -0.5 i \epsilon(t)^* \exp(i \omega t)  \exp(-i \phi))

        The 0.5 are due to the Wirtinger derivatives due to Re part.

        Parameters
        ----------
        time: Array
            Single timestamp.

        Returns
        -------
        Array
            Return the gradients from all devices at one time.

        """
        phase_fac = jnp.exp(-1j * self._phase.get_value())
        lo_out = jnp.squeeze(self._lo.get_value(time), axis=0)
        sig = self._complex_signal(time)
        gradients = jnp.zeros(shape=(0,))

        # Collect gradients for envelopes
        for dev in self._envs:
            _, grad = dev.get_value_and_gradient(time)
            grad = jnp.squeeze(grad.conj(), axis=0)
            if grad.size != 0:
                grad *= lo_out * phase_fac
            gradients = jnp.append(gradients, 0.5 * jnp.real(grad), axis=0)

        # Collect LO gradients
        lo_freq = self._lo.get_parameters()[0]
        if self._is_optimized(lo_freq):
            gradients = jnp.append(gradients, 0.5j * time * sig, axis=0)

        # Collect gradient of Phase
        if self._is_optimized(self._phase):
            gradients = jnp.append(
                gradients,
                -0.5j * sig,
                axis=0,
            )
        return gradients



"""Test the composite Hamiltonian model."""

import numpy as np
import pytest

from paraqeet.model.composite_hamiltonian import CompositeHamiltonian
from paraqeet.model.coupling import TwoBodyCoupling
from paraqeet.model.drive_operator import DriveOperator
from paraqeet.model.transmon import Transmon
from paraqeet.quantity import Quantity
from paraqeet.signal.envelopes import FlatTopGaussianEnvelope
from paraqeet.signal.iq_mixer import IQMixer

LEN_SIG = 101


@pytest.fixture
def time_samples():
    """Generate time samples according to the given signal length.

    The given signal length `LEN_SIG` is a module level global variable.

    """
    return np.linspace(0, 10e-9, LEN_SIG)


@pytest.fixture
def tone():
    """Return a cosine tone."""
    tone = FlatTopGaussianEnvelope()
    tone.set_optimizable_parameters(tone.get_parameters())
    return tone


@pytest.fixture
def gen(tone):
    """Return a sinusoidal generator object."""
    gen = IQMixer(envelopes=[tone])
    return gen


@pytest.fixture
def drive(gen):
    """Return a generator drive object."""
    drive = DriveOperator(gen, is_longitudinal=False)
    return drive


@pytest.fixture
def transmon_parameters():
    """Return a random parameter generating function for transmons."""

    class RandomParameters:
        def get(self):
            freq = np.random.uniform(5.5, 6.0) * 1e9 * 2 * np.pi
            anharm = -np.random.uniform(2.0, 2.4) * 1e7 * 2 * np.pi
            return (freq, anharm)

    return RandomParameters()


@pytest.fixture
def transmon(transmon_parameters, drive):
    """Return a transmon generating function."""

    def get(dimension):
        freq, anharm = transmon_parameters.get()
        transmon = Transmon(
            dimension=dimension,
            frequency=Quantity(freq, 0.8 * freq, 1.2 * freq),
            anharmonicity=Quantity(anharm, 1.2 * anharm, 0.8 * anharm),
            drives=[drive],
        )
        return transmon

    return get


@pytest.fixture
def uncoupled_transmons(transmon):
    """Return a composite Hamiltonian generating function."""

    def _method(dim1, dim2):
        transmon1 = transmon(dim1)
        transmon2 = transmon(dim2)
        compositeHams = CompositeHamiltonian([transmon1, transmon2])
        return compositeHams

    return _method


@pytest.fixture
def coupled_transmons(transmon):
    """Return a coupled transmon generating function."""

    def _method(dim1: int, dim2: int, use_rwa: bool = False):
        transmon1 = transmon(dim1)
        transmon2 = transmon(dim2)

        couplingStr = np.abs(transmon1.frequency.get_value() - transmon2.frequency.get_value()) * 0.05
        coupling = TwoBodyCoupling(
            transmon1,
            transmon2,
            is_longitudinal=False,
            coefficient=Quantity(couplingStr, 0.8 * couplingStr, 1.2 * couplingStr, "Hz"),
            use_rwa=use_rwa,
        )

        compositeHams = CompositeHamiltonian([transmon1, transmon2], [coupling])
        return compositeHams

    return _method


@pytest.fixture
def coupled_transmons_chain(transmon, random_quantity):
    def _method(dims: list[int], open_system: bool = True) -> CompositeHamiltonian:
        transmons = [transmon(d) for d in dims]
        if open_system:
            for t in transmons:
                t.temp = random_quantity(1, "K")
                t.t1 = random_quantity(1, "")
                t.t2star = random_quantity(1, "")
        coupling_strengths = [
            0.5 * np.abs(transmons[i].frequency.get_value() - transmons[i + 1].frequency.get_value())
            for i in range(len(transmons) - 1)
        ]
        couplings = [
            TwoBodyCoupling(
                subsystem_A=transmons[i],
                subsystem_B=transmons[i + 1],
                is_longitudinal=False,
                coefficient=Quantity(
                    coupling_strengths[i], 0.8 * coupling_strengths[i], 1.2 * coupling_strengths[i], "Hz"
                ),
            )
            for i in range(len(transmons) - 1)
        ]

        return CompositeHamiltonian(transmons, couplings)

    return _method


def test_dimension(coupled_transmons):
    """Test dimension of matrix produced by compositeHamiltonian."""
    for _ in np.arange(1, 10):
        dim1 = np.random.randint(2, 6)
        dim2 = np.random.randint(2, 7)
        hamil = coupled_transmons(dim1, dim2)
        assert hamil.dimension() == dim1 * dim2


def test_get_value_one_time(uncoupled_transmons):
    """Test shape of Matrix produced by compositeHamiltonian."""
    for _ in np.arange(1, 10):
        dim1 = np.random.randint(2, 6)
        dim2 = np.random.randint(2, 7)
        hamil = uncoupled_transmons(dim1, dim2)
        hams = hamil.get_value_at_timestep(0)
        assert hams.shape == (dim1 * dim2, dim1 * dim2)


def test_get_value_one_time_rwa(coupled_transmons):
    """Test shape of Matrix produced by compositeHamiltonian."""
    for _ in np.arange(1, 10):
        dim1 = np.random.randint(2, 6)
        dim2 = np.random.randint(2, 7)
        hamil = coupled_transmons(dim1, dim2, use_rwa=True)
        hams = hamil.get_value_at_timestep(0)
        assert hams.shape == (dim1 * dim2, dim1 * dim2)


def test_get_value(coupled_transmons, time_samples):
    """Test shape of Matrix produced by compositeHamiltonian."""
    for _ in np.arange(1, 10):
        dim1 = np.random.randint(2, 6)
        dim2 = np.random.randint(2, 7)
        hamil = coupled_transmons(dim1, dim2)
        hams = hamil.get_value(time_samples)
        assert hams.shape == time_samples.shape + (dim1 * dim2, dim1 * dim2)


def test_gradient(gen, coupled_transmons, time_samples):
    """Test shape of gradients by compositeHamiltonian.

    Number of gradient parameters include gradients from both the drives, and
    both the transmon frequency, anharmonicity and the coupling.
    """
    for _ in np.arange(1, 5):
        dim1 = np.random.randint(2, 6)
        dim2 = np.random.randint(2, 7)
        hamil = coupled_transmons(dim1, dim2)
        hamil.set_optimizable_parameters(hamil.get_parameters())
        _, grads = gen.get_value_and_gradient(time_samples)
        _, ham_grads = hamil.get_value_and_gradient(time_samples)
        assert ham_grads.shape == (
            grads.shape[0],
            grads.shape[1] * 2 + 5,
            dim1 * dim2,
            dim1 * dim2,
        )


def test_collapseops_dimensions(coupled_transmons_chain):
    for _ in np.arange(1, 10):
        dims = [np.random.randint(2, 5) for _ in range(np.random.randint(2, 5))]
        hamil = coupled_transmons_chain(dims, True)
        ops = hamil.get_collapseops()
        for rate, op in ops:
            assert op.shape[0] == op.shape[1] == hamil.dimension()



"""Class definition of the Weighted Sum Goal model."""

import jax.numpy as jnp
import numpy as np

from paraqeet.differentiable import Differentiable
from paraqeet.measurement.measurement import NormalizableMeasurement
from paraqeet.measurement.state_transfer_fidelity import (
    StateTransferFidelityGRAPE,
)
from paraqeet.propagation.differentiable_propagation import DifferentiablePropagation
from paraqeet.quantity import Array
from paraqeet.signal.pwc_generator import PWCGenerator


class GOATOverGRAPE(NormalizableMeasurement, Differentiable):
    """Combine GRAPE propagation with analytic gradients of GOAT via chain rule.

    Parameters
    ----------
    measurement : StateTransferFidelityGRAPE
        A StateTransferFidelityGRAPE measurement.
    propagation: DifferentiablePropagation
        Propagation method used for the optimization. Used to determine the time grid.
    generators: PWCGenerator | list[PWCGenerator]
        A PWCGenerator or a list of PWCGenerators that are used for propagation.
    """

    _measurement: StateTransferFidelityGRAPE
    _gens: list[PWCGenerator]
    _propagation: DifferentiablePropagation

    def __init__(
        self,
        measurement: StateTransferFidelityGRAPE,
        propagation: DifferentiablePropagation,
        generators: PWCGenerator | list[PWCGenerator],
    ):
        self._measurement = measurement
        self._propagation = propagation
        self._gens = generators if isinstance(generators, list) else [generators]
        for gen in self._gens:
            gen.set_optimizable_parameters(gen.get_parameters())

    def _pad_with_zeros(self, grad: Array, gen_num: int) -> Array:
        """Pad gradient with zeros depending on the subsystem number and number of PWC pixels in the pulses.

        Parameters
        ----------
        grad : Array
            Gradient from a subsystem.
        gen_num : int
            Subsystem number, also determined by the generator order.

        Returns
        -------
        Array
            Return padded gradient vector.
        """
        num_pixels, num_params = grad.shape
        padded_grad = np.zeros((0, num_params))
        for i in range(len(self._gens)):
            if i == gen_num:
                padded_grad = np.append(padded_grad, grad, axis=0)
            else:
                padded_grad = np.append(padded_grad, np.zeros((num_pixels, num_params)), axis=0)
        return jnp.array(padded_grad)

    def measure(self, times: Array) -> Array | float:
        """Sum of plain weighted measurements.

        Returns
        -------
        Array
            Returns the plain weighted sum.

        """
        grape = self._measurement
        for gen in self._gens:
            gen._update_inphase_and_outofphase()
        return grape.measure(times=times)

    def calculate_normalized_scalar(self, times: Array | float) -> float:
        """Passthrough the measurement.

        Returns
        -------
        Array
            Returns the normalized weighted sum.

        """
        grape = self._measurement
        for gen in self._gens:
            gen._update_inphase_and_outofphase()
        return grape.calculate_normalized_scalar(times=times)

    def get_value_and_gradient(self, times: Array) -> tuple[Array, Array] | tuple[float, Array]:
        """Compute gradients with GRAPE and use the chain rule
        to provide the gradients for the optimizer.

        Returns
        -------
        Array
            Function value.
        Array
            Gradients.

        """
        # TODO: Fix typing
        grape = self._measurement
        for gen in self._gens:
            gen._update_inphase_and_outofphase()

        # Construct the same time grid as propagation to evaluate control gradients
        interp_times = jnp.array([])
        for ti in range(1, len(times)):
            t_interpolated, dt = self._propagation._construct_times(times, ti)
            interp_times = jnp.append(interp_times, t_interpolated, axis=0)

        time_grid = interp_times[:-1] + dt / 2

        control_gradients: list[Array] = []
        for gen_num, gen in enumerate(self._gens):
            control_gradients.append(self._pad_with_zeros(gen._get_partial_derivatives(time_grid), gen_num))
        control_gradients_arr = jnp.hstack(control_gradients)

        # Evaluate GRAPE gradients
        function_value, grape_gradients = grape.get_value_and_gradient(interp_times)

        # Reconstruct GOAT gradients
        goat_gradients = control_gradients_arr.T @ grape_gradients
        return function_value, goat_gradients



"""Test the functionality of the generation of the DRAG corrected signal."""

import numpy as np
import pytest

from paraqeet.signal.envelopes import (
    FlatTopGaussianEnvelope,
    GaussEnvelope,
    ZeroEnvelope,
)
from paraqeet.signal.iq_mixer import IQMixer
from paraqeet.signal.waveform import DRAGMixer

LEN_SIG = 1001


@pytest.fixture
def time_samples():
    """Generate time samples from the given signal length.

    The given signal length `LEN_SIG` is a module level global variable.

    """
    return np.linspace(0, 10e-9, LEN_SIG)


@pytest.fixture
def gen():
    """Return DRAG signal generator object."""
    tone = GaussEnvelope()
    drag_tone = DRAGMixer(tone)
    return IQMixer(envelopes=[drag_tone])


@pytest.fixture
def zero():
    """Return a zero signal device."""
    return ZeroEnvelope()


@pytest.fixture
def gauss():
    """Return a gaussian signal device."""
    return GaussEnvelope()


@pytest.fixture
def zero_gen():
    """Return a zero tone DRAG generator object."""
    tone = ZeroEnvelope()
    drag_tone = DRAGMixer(tone)
    return IQMixer(envelopes=[drag_tone])


@pytest.fixture
def flattop():
    """Return a FlatTop signal with DRAG."""
    tone = FlatTopGaussianEnvelope()
    drag_tone = DRAGMixer(tone)
    return IQMixer(envelopes=[drag_tone])


@pytest.fixture
def gen_multiple_tones():
    """Return a multiple tone DRAG generator."""
    tone1 = GaussEnvelope()
    params1 = tone1.get_parameters()
    drag_tone1 = DRAGMixer(tone1)

    tone2 = GaussEnvelope()
    params2 = tone2.get_parameters()
    drag_tone2 = DRAGMixer(tone2)
    return (
        IQMixer(envelopes=[drag_tone1, drag_tone2]),
        np.concatenate((params1, params2)),
    )


def test_constant_env(zero_gen, time_samples):
    """Test the values of a DRAG signal using a constant envelope."""
    assert np.all(zero_gen.get_value(time_samples) == np.zeros_like(time_samples))


def test_gen(gen, time_samples) -> None:
    """Computes a sample signal and checks vectorized generation."""
    sig = gen.get_value(time_samples)
    assert len(sig) == LEN_SIG


def test_get_parameters(gen_multiple_tones):
    """Test if the expected amount of parameters is present.

    If multiple tones define the total envelope, the signal needs to have the
    sum of all parameters of the envelope tones plus 4 parameters.

    """
    gen, all_params = gen_multiple_tones
    params = gen.get_parameters()
    assert len(params) == len(all_params) + 4


def test_gradient_shape(gen, time_samples):
    """Test the length of the signal gradient."""
    gen.set_optimizable_parameters(gen.get_parameters())
    _, grads = gen.get_value_and_gradient(time_samples)
    assert grads.shape[0] == time_samples.shape[0]


def test_gradient_flattop(flattop, time_samples):
    """Test the length of the gradient of flattop signal."""
    flattop.set_optimizable_parameters(flattop.get_parameters())
    _, grads = flattop.get_value_and_gradient(time_samples)
    assert grads.shape[0] == time_samples.shape[0]



import numpy as np
import pytest

from paraqeet.model.open_system import OpenSystem
from paraqeet.model.resonator import Resonator
from paraqeet.quantity import Quantity


@pytest.fixture
def hamiltonian():
    def _method(dimension):
        FREQ = 4.8e9 * 2 * np.pi
        return Resonator(
            frequency=Quantity(FREQ, 0.8 * FREQ, 1.2 * FREQ),
            dimension=dimension,
        )

    return _method


@pytest.fixture
def open_system(
    hamiltonian,
):
    def _method(dimension):
        hamil = hamiltonian(dimension)
        hamil.t1 = Quantity(1e-9, 1e-9, 100e-6)
        hamil.temp = Quantity(10e-3, 1e-3, 50e-3)
        hamil.t2star = Quantity(10e-9, 1e-9, 100e-6)
        return OpenSystem(hamil)

    return _method


def test_create_dense_matrix(open_system):
    for dim in range(3, 6):
        random_time_vector = np.linspace(0.0, np.random.randint(1, 10) * np.random.rand(), np.random.randint(1, 10))

        system = open_system(dim)
        system.sparse_superop = True
        assert system.sparse_superop is True
        matrix = system.get_value(random_time_vector)
        assert matrix.shape == (len(random_time_vector), dim**2, dim**2)


def test_create_sparse_matrix(open_system):
    for dim in range(3, 6):
        random_time_vector = np.linspace(0.0, np.random.randint(1, 10) * np.random.rand(), np.random.randint(1, 10))

        system = open_system(dim)
        system.sparse_superop = False
        assert system.sparse_superop is False
        matrix = system.get_value(random_time_vector)
        assert matrix.shape == (len(random_time_vector), dim**2, dim**2)


def test_has_parameters(open_system):
    system = open_system(np.random.randint(2, 10))
    assert system.get_parameters() is not None
    assert len(system.get_parameters()) >= 0



"""Test the Resonator model."""

import numpy as np
import pytest

from paraqeet.differentiable import Differentiable
from paraqeet.model.drive_operator import DriveOperator
from paraqeet.model.open_system import OpenSystem
from paraqeet.model.resonator import Resonator
from paraqeet.propagation.scipy_expm import ScipyExpm
from paraqeet.propagation.vern7 import Vern7
from paraqeet.quantity import Quantity
from paraqeet.signal.envelopes import FlatTopGaussianEnvelope, ZeroEnvelope
from paraqeet.signal.iq_mixer import IQMixer

DIMS = 3
FREQ = 4.8e9 * 2 * np.pi
LEN_SIG = 1001
T1 = Quantity(1e-9, 1e-9, 100e-6)
TEMP = Quantity(10e-3, 1e-3, 50e-3)
T2STAR = Quantity(10e-9, 1e-9, 100e-6)


@pytest.fixture
def time_samples():
    """Generate time samples from the given signal length.

    The given signal length `LEN_SIG` is a module level global variable.

    """
    return np.linspace(0, 10e-9, LEN_SIG)


@pytest.fixture
def tone():
    """Return a object for sinusoidal tone generator with error envelopes."""
    return FlatTopGaussianEnvelope()


@pytest.fixture
def gen(tone):
    """Return a sinusoidal tone generator object."""
    gen = IQMixer(envelopes=[tone])
    return gen


@pytest.fixture
def hamiltonian(gen):
    """Return a resonator object."""

    def _method(dimension):
        drive = DriveOperator(gen, is_longitudinal=False)
        return Resonator(
            dimension=dimension,
            frequency=Quantity(FREQ, 0.8 * FREQ, 1.2 * FREQ),
            drives=[drive],
        )

    return _method


@pytest.fixture
def open_resonator():
    """Return an open model for the resonator."""
    tone = ZeroEnvelope()
    generator = IQMixer(envelopes=[tone])
    drive = DriveOperator(generator, is_longitudinal=False)
    resonator = Resonator(
        frequency=Quantity(FREQ, 0.8 * FREQ, 1.2 * FREQ),
        drives=[drive],
        dimension=DIMS,
    )
    resonator.t1 = T1
    resonator.temp = TEMP
    resonator.t2star = T2STAR
    model = OpenSystem(resonator)

    return model


@pytest.fixture
def expm(open_resonator):
    init = np.zeros((DIMS, 1), dtype=np.complex128)
    init[DIMS - 1][0] = 1  # Fully excited state
    init_dm = np.matmul(init, init.T)

    prop = ScipyExpm(open_resonator, resolution=100e9)
    prop.set_initial_state(init_dm)
    return prop


@pytest.fixture
def ode(open_resonator):
    init = np.zeros((DIMS, 1), dtype=np.complex128)
    init[DIMS - 1][0] = 1  # Fully excited state
    init_dm = np.matmul(init, init.T)

    open_resonator.ode_propagation = True
    prop = Vern7(open_resonator, resolution=100e9)
    prop.set_initial_state(init_dm)
    return prop


def test_get_value(hamiltonian, time_samples):
    """Test the getMatrix method."""
    for dim in np.arange(1, 10):
        hamil = hamiltonian(dim)
        hams = hamil.get_value(time_samples)
        assert hams.shape == time_samples.shape + (dim, dim)


def test_gradient(gen, hamiltonian, time_samples):
    """Test the gradients of the system.

    The Hamiltonian should have all derivatives of the drive
    plus the derivative w.r.t. the resonator frequency.

    """
    for dim in np.arange(1, 10):
        hamil = hamiltonian(dim)
        if not isinstance(hamil, Differentiable):
            continue
        hamil.set_optimizable_parameters(hamil.get_parameters())
        _, grads = gen.get_value_and_gradient(time_samples)
        #  TODO: fix error related to the length of time_samples-array
        _, ham_grads = hamil.get_value_and_gradient(time_samples)
        assert ham_grads.shape == (grads.shape[0], grads.shape[1] + 1, dim, dim)


def test_decay_expm(expm):
    t_final = 20e-9
    ts = np.linspace(0, t_final, 101)

    states = expm.propagate(ts)
    final_state = states[-1]

    # Check if final state is density matrix
    assert np.isclose(np.linalg.trace(final_state), 1)

    # Check the excited state population
    assert np.isclose(final_state[DIMS - 1, DIMS - 1], 0)

    # Check the ground state population
    assert np.isclose(final_state[0, 0], 1)


def test_decay_ode(ode):
    t_final = 20e-9
    ts = np.linspace(0, t_final, 101)

    states = ode.propagate(ts)
    final_state = states[-1]

    # Check if final state is density matrix
    assert np.isclose(np.linalg.trace(final_state), 1)

    # Check the excited state population
    assert np.isclose(final_state[DIMS - 1, DIMS - 1], 0)

    # Check the ground state population
    assert np.isclose(final_state[0, 0], 1)



"""Test dCRAB optimization using GOAT over GRAPE. This is same as the example 02E_GOAToverGRAPE_dCRAB"""

import numpy as np
import pytest

from paraqeet.measurement.goat_over_grape import GOATOverGRAPE
from paraqeet.measurement.state_transfer_fidelity import StateTransferFidelityGRAPE
from paraqeet.model.closed_system import ClosedSystem
from paraqeet.model.rotating_frame_drive import RotatingFrameDrive
from paraqeet.optimization_map import OptimizationMap
from paraqeet.optimizers.dcrab_optimizer_gradient import DCRABOptimizerGradient
from paraqeet.propagation.scipy_expm_grape import ScipyExpmGRAPE
from paraqeet.quantity import Quantity
from paraqeet.signal.envelopes import DCRABEnvelope
from paraqeet.signal.pwc_generator import PWCGenerator
from paraqeet.signal.waveform import FlatTopGaussianFilter
from tests.model.spin_rwa import SpinRWA

T_FINAL = 20e-9
TLIST = np.linspace(0, T_FINAL, 40)
EPS = 5 * np.pi * 1.0  # initial amplitude of the resonator (in MHz)
EPS_MAX = 10 * EPS  # maximum amplitude of the resonator (in MHz)


@pytest.fixture
def tone():
    tone = DCRABEnvelope(
        num_components=2,
        max_frequency=5 * 2 * np.pi,
        amplitude=Quantity(EPS * 1e6, -EPS_MAX * 1e6, EPS_MAX * 1e6, name="Amplitude"),
        t_final=Quantity(T_FINAL, 1 / 2 * T_FINAL, 2 * T_FINAL, name="t_final"),
        seed=18537,
    )
    return tone


@pytest.fixture
def gen(tone):
    smooth_tone = FlatTopGaussianFilter(
        envelopes=tone, t_final=Quantity(T_FINAL, 1 / 2 * T_FINAL, 2 * T_FINAL, name="t_final")
    )
    gen = PWCGenerator(envelopes=[smooth_tone], tlist=TLIST, max_amplitude=np.sqrt(2) * EPS_MAX * 1e6)
    return gen


@pytest.fixture
def model(gen):
    drive = RotatingFrameDrive(gen)
    spin = SpinRWA(drives=[drive])
    model = ClosedSystem(spin)
    return model


@pytest.fixture
def prop(model):
    prop = ScipyExpmGRAPE(model, resolution=1e9)

    init = np.array([[1.0], [0]])  # |0>
    target = np.array([[0.0], [1]])  # |1>

    prop.set_initial_state(init)
    prop.set_target_state(target)
    return prop


@pytest.fixture
def fid(prop):
    init = np.array([[1.0], [0]])  # |0>
    target = np.array([[0.0], [1]])  # |1>

    zeroone = StateTransferFidelityGRAPE(
        propagation=prop,
        initial_state=init,
        target_state=target,
    )
    return zeroone


@pytest.fixture
def opt_grad(tone, fid, gen, prop):
    optmap = OptimizationMap()
    params = tone.get_parameters()
    optmap.add(tone, [params[0]] + params[2:])
    optmap.register_params_with_optimizables()

    goat = GOATOverGRAPE(fid, prop, generators=[gen])
    opt_grad = DCRABOptimizerGradient(
        goat,
        optimization_map=optmap,
        super_iteration_every=150,
        max_super_iteration_num=5,
        print_every_iteration_num=10,
        super_iteration_tol=1e-9,
        seed=19573,
    )
    return opt_grad


def test_goat_over_grape(opt_grad):
    res = opt_grad.optimize(TLIST)
    assert res.value < 1e-4



"""Class definition of JSONFileSerializer."""

import json

from paraqeet.exceptions import SerializationException
from paraqeet.serialization.serializer import Serializer


class JSONFileSerializer(Serializer):
    """Writes data into and read data from JSON files in a human-readable format."""

    _COMMENT_KEY = "_comment"
    _file: str

    def __init__(self, file: str):
        self._file = file

    def save(self, data: dict, comment: str | None = None) -> None:
        """Saves the data and the optional comment to the JSON file that was specified in the constructor."""
        # The comment is simply stored in the same dict
        if comment:
            data[self._COMMENT_KEY] = comment
        with open(self._file, "w", encoding="utf-8") as f:
            json.dump(data, f)

    def load(self) -> dict:
        """Loads and returns the data from JSON file"""
        with open(self._file) as f:
            data = json.load(f)
            if not isinstance(data, dict):
                raise SerializationException("File does not contain a dictionary.")

            if self._COMMENT_KEY in data:
                del data[self._COMMENT_KEY]
            return data

    def load_comment(self) -> str | None:
        """Loads and returns the comment from the JSON file."""
        with open(self._file) as f:
            data = json.load(f)
            return data[self._COMMENT_KEY] if self._COMMENT_KEY in data else None



"""Class definition of a Drive optimizable model."""

from abc import abstractmethod

from jax import vmap

from paraqeet.optimizable import Optimizable
from paraqeet.quantity import Array


# TODO: Is Drive a Differentiable object? If so, it should inherit from Differentiable at least 
# for the purpose of clarity and consistency. Would it make sense to add a default 
# implementation of the abstract Differentiable method get_value_and_gradient?
# If not, we should rename the methods to e.g. get_hamiltonian_gradient to avoid confusion.
class Drive(Optimizable):
    """Represents a time-dependent drive on a subsystem.

    This can for example be a microwave or flux drive.

    """

    def get_value(self, annihilation_operator: Array, times: Array) -> Array:
        """Return the matrix representation of the drive.

        The dimension is given by the Hamiltonian to which this drive is
        attached. The default implementation calls getMatrixOneTime for each
        time step. Subclasses can override this function for a more efficient
        implementation.

        Parameters
        ----------
        annihilation_operator : Array
            Operator of the subsystem to which this drive is attached
        times: Array
            Vector of time samples.

        Returns
        -------
        Array
            Matrix of shape [t, n, n]  with 't' as time and 'n' as the Hilbert
            space dimension.

        """
        # vmap iterates over the times array and returns float. Not caught by mypy.
        return vmap(self.get_value_at_timestep, in_axes=(None, 0))(annihilation_operator, times)  # type: ignore

    @abstractmethod
    def get_value_at_timestep(self, annihilation_operator: Array, t: float) -> Array:
        """Return the matrix representation of the drive.

        The dimension is given by the Hamiltonian to which this drive is
        attached.

        Parameters
        ----------
        annihilation_operator
            Operator of the subsystem to which this drive is attached.
        t: float
            One time point.

        Returns
        -------
        Array
            Matrix of shape [n, n]  with `n` as the Hilbert space dimension.

        """
        pass

    def get_gradient(self, annihilation_operator: Array, times: Array) -> Array:
        """Return the gradient of the system.

        Returns the gradient of the matrix representation of the Hamiltonian
        with respect to each parameter as a list.

        Parameters
        ----------
        annihilation_operator : Array
            Operator of the subsystem to which this drive is attached.
        times: Array
            Vector of time samples.

        Returns
        -------
        Array
            Array of shape [t, p, n, n] with 't' as time, 'p' as number of
            parameters and 'n' as the Hilbert space dimension.


        """
        # Ignoring mypy here as vmap makes the array to float
        return vmap(self.get_gradient_at_timestep, in_axes=(None, 0))(annihilation_operator, times)  #  type: ignore

    @abstractmethod
    def get_gradient_at_timestep(self, annihilation_operator: Array, timestep: float) -> Array:
        """Get the one-time gradient of the system.

        Returns the gradient of the matrix representation of the
        Hamiltonian with respect to each parameter as a list.

        Parameters
        ----------
        annihilation_operator : Array
            Operator of the subsystem to which this drive is attached.
        timestep: float
            One time step.

        Returns
        -------
        Array
            Array of shape [p, n, n] with 'p' as the number
            of parameters and 'n' as the  Hilbert space dimension.

        """
        pass

    @staticmethod
    def _repeat(mat: Array, num: int) -> Array:
        """Repeats the matrix mat for each timestep in the times array.

        Returns an array with shape [t, n, m] where 't' is the
        number of time steps and 'mat' is an 'n' times 'm' matrix.

        Parameters
        ----------
        mat: Array
            Input matrix for repetition.
        num : int
            Number of times of repetition.

        Returns
        -------
        Array
            Repeated matrix for further computation.

        """
        return mat.reshape((1,) + mat.shape).repeat(num, axis=0)



"""Class definition of an open system."""

from collections.abc import Callable

import jax.numpy as jnp
from jax import jit, vmap
from jax.experimental.sparse import BCOO

from paraqeet.differentiable import Differentiable
from paraqeet.model.differentiable_hamiltonian import DifferentiableHamiltonian
from paraqeet.model.equation_of_motion import EquationOfMotion
from paraqeet.quantity import Array, Quantity


class OpenSystem(EquationOfMotion):
    """
    Model of an open quantum system, defined by the Hamiltonian and collapse operators.
    Its dynamics given by the Lindblad master equation.

    Currently the gradients for ODE propagation methods is not supported.

    Parameters
    ----------
    hamiltonian : Hamiltonian
        Matrix representation of a Hamiltonian.
    sparse_superop: bool
        Flag to save superoperator as sparse matrices.
    ode_propagation: bool
        Flag to use ODE methods for propagation.
        If `true` then `get_value` method returns list of Hamiltonian (with time) and collapse operator.
        Else returns Lindblad superoperator.
    """

    _ode_propagation: bool
    _sparse_superop: bool
    _get_value_method: Callable
    _hamiltonian: DifferentiableHamiltonian

    def __init__(
        self,
        hamiltonian: DifferentiableHamiltonian,
        sparse_superop: bool = False,
        ode_propagation: bool = False,
    ):
        self._hamiltonian = hamiltonian
        self._sparse_superop = sparse_superop
        self.ode_propagation = ode_propagation

    @property
    def sparse_superop(self) -> bool:
        """Flag to store superoperators as sparse matrices.

        Returns
        -------
        bool
            Flag to store sparse matrices.
        """
        return self._sparse_superop

    @sparse_superop.setter
    def sparse_superop(self, sparse_superop: bool) -> None:
        self._sparse_superop = sparse_superop

    @property
    def ode_propagation(self) -> bool:
        """Flag to set method of propagation to ODE.

        Returns
        -------
        bool
            Flag to use ODE propagation.
        """
        return self._ode_propagation

    @ode_propagation.setter
    def ode_propagation(self, ode_propagation: bool) -> None:
        self._ode_propagation = ode_propagation

        if ode_propagation:
            self._get_value_method = self._get_ode_propagation_eom
        else:
            self._get_value_method = vmap(self._create_lindbladian_superop)

    def get_parameters(self) -> list[Quantity]:
        """Get a list of optimizable parameters.

        Returns
        -------
        list[Quantity]
            list of optimizable parameters of the system.

        """
        return self._hamiltonian.get_parameters()

    def get_collapseops(self) -> list[tuple[Array, Array]]:
        """Get a list of tuples of decay rates and collapse operators for each subsystem.

        Returns
        -------
        list[tuple[float, Array]]
            list of collapse operators

        """
        return self._hamiltonian.get_collapseops()

    def _get_ode_propagation_eom(self, times: Array) -> tuple[Array, list[Array]]:
        """
        Return the coherent and incoherent EOM parts separately.
        Here the coherent part is the Hamiltonian as a function of time (w/o -1j)
        and the incoherent part is a list of collapse operators

        Parameters
        ----------
        times: Array
            Vector of time samples

        Returns
        -------
        tuple[Array, Array]
             Hamiltonian EOM ([t, N, N] matrix) and the `m` collapse operators ([m, N^2, N^2] matrix)
        """
        ham_eom = self._hamiltonian.get_value(times)
        rates_and_cols = self.get_collapseops()
        cols: list[Array] = [jnp.sqrt(rate) * col for rate, col in rates_and_cols]
        return -1j * ham_eom, cols

    def _create_hamiltonian_superop(self, t) -> Array | BCOO:
        """Create the Hamiltonian superoperator for one time point `t`."""
        identityop = jnp.eye(self._hamiltonian.dimension())
        ham = self._hamiltonian.get_value_at_timestep(t)
        superop = -1j * jnp.kron(identityop, ham) + 1j * jnp.kron(ham.T, identityop)
        if self.sparse_superop:
            return BCOO.fromdense(superop)
        return superop

    def _create_collapse_superop(self) -> Array | BCOO:
        """Create the superoperator due to the collapse part. This is time independent."""
        dim = self._hamiltonian.dimension()
        identityop = jnp.eye(dim)
        superop = jnp.zeros((dim**2, dim**2), dtype=jnp.float64)
        rates_and_cols = self._hamiltonian.get_collapseops()
        for rate, col in rates_and_cols:
            superop += rate * jnp.kron(col.conj(), col)
            superop -= rate * jnp.kron(jnp.matmul(col.T, col.conj()), identityop) / 2
            superop -= rate * jnp.kron(identityop, jnp.matmul(col.conj().T, col)) / 2

        if self.sparse_superop:
            return BCOO.fromdense(superop)
        return superop

    def _create_lindbladian_superop(self, t) -> Array | BCOO:
        """Create the Lindbladian superoperator for one time point `t`."""
        ham_super_op = self._create_hamiltonian_superop(t)
        col_super_op = self._create_collapse_superop()
        return ham_super_op + col_super_op

    # TODO: check the times-Array: internally a method might be called which expects only one timestep
    def get_value(self, times: Array):
        """
        Computes the right hand side of the Schrödinger equation without multiplying the state.
        Used for unitary solvers.

        Parameters
        ----------
        times: Array
            Vector of time samples

        Returns
        -------
        Array
            RHS with dimension [t, n, n]  with t: time, n: hilbert space
        """
        # TODO: in case the matrix_method is _create_lindbladian_superop, only one timestep is expected!
        return self._get_value_method(times)

    @staticmethod
    @jit
    def _kron(A, B):
        return jnp.kron(A, B)

    def _create_hamiltonian_grad_superop(self, timestep: float):
        """Create the Gradient of Hamiltonian superoperator for one time point `timestep`."""
        identityop = jnp.eye(self._hamiltonian.dimension())
        ham_grad = self._hamiltonian.get_gradient_at_timestep(timestep)
        term1 = -1j * vmap(OpenSystem._kron, in_axes=(None, 0))(identityop, ham_grad)
        term2 = 1j * vmap(OpenSystem._kron, in_axes=(0, None))(jnp.transpose(ham_grad, axes=(0, 2, 1)), identityop)
        superop = term1 + term2
        return superop

    def get_value_and_gradient(self, times: Array) -> tuple[Array, Array]:
        """Compute the gradient of get_value."""
        # TODO: What to do if Hamiltonian is not differentiable?
        if isinstance(self._hamiltonian, Differentiable):
            if self.ode_propagation:
                # TODO: Can the eom be obtained without calling the get_value method?
                _, grads = self._hamiltonian.get_value_and_gradient(times)
                eom = self._get_value_method(times)
                grads = -1j * grads
            else:
                eom = vmap(self._create_lindbladian_superop)(times)
                # TODO: times is an Array but float is expected
                # ignoring mypy due to vmap
                grads = vmap(self._create_hamiltonian_grad_superop)(times)  # type: ignore
        return eom, grads



"""Class definition of the unitary fidelity model."""

import jax
import jax.numpy as jnp

from paraqeet.differentiable import Differentiable
from paraqeet.measurement.measurement import NormalizableMeasurement
from paraqeet.propagation.differentiable_propagation import DifferentiablePropagation
from paraqeet.quantity import Array

jax.config.update("jax_enable_x64", True)


class UnitaryFidelity(NormalizableMeasurement, Differentiable):
    """Unitary fidelity measurement model.

    Fidelity measure that compares the propagator with a desired gate
    by way of L2 norm.

    Parameters
    ----------
    propagation : Propagation
        Implementation of EOM solver.
    gate : Array
        Matrix representation of target gate.
    times : Array
        List of times to compare. Should have length 2.
        More is allowed, but only the first and last are used.
    basis_states : Array optional
        List of basis states.
        If set the ideal and actual gate are applied to these states
        and their pairwise overlap computed, equivalent to the L2 trace norm.
        Defaults to [].

    """

    _basis_states: Array | None
    _target_costates: Array
    _propagation: DifferentiablePropagation

    def __init__(
        self,
        propagation: DifferentiablePropagation,
        gate: Array,
        basis_states: Array | None = None,
    ):
        self._propagation = propagation
        if basis_states is not None:
            self._propagation.set_initial_state(basis_states)
        else:
            basis_states = jnp.eye(gate.shape[0])
        self._basis_states = basis_states
        self.set_ideal_gate(gate)

    # TODO: since this method is declared as static, it belongs to the class, not to the instance. 
    # It should be called accordingly.
    @staticmethod
    def _fid(overlaps: Array) -> float:
        """Gate fidelity from state overlaps.

        Parameters
        ----------
        overlaps : List
            State overlap as a one-dimensional array.

        Returns
        -------
        float
            Gate fidelity as a single float.

        """
        return float(jnp.abs(jnp.average(overlaps)) ** 2)

    def measure(self, times: Array) -> Array | float:
        """Return measurement in the range [0, 1]."""
        return self.calculate_normalized_scalar(times=times)

    def calculate_normalized_scalar(self, times: Array | float) -> float:
        """Return the L2 norm of the last time step compared to the ideal gate.

        Returns
        -------
        Array
            L2 norm of the last time step compared to the ideal gate.

        """
        # TODO: Fix typing
        states = self._propagation.propagate(time=times)
        overlaps = []
        for ii, s in enumerate(self._target_costates.T):
            overlaps.append(jnp.vdot(s, states[-1][:, ii]))
        return UnitaryFidelity._fid(jnp.asarray(overlaps))

    def get_value_and_gradient(self, times: Array) -> tuple[Array, Array] | tuple[float, Array]:
        """Get the L2 norm and the analytic expression for the gradient.

        Returns
        -------
        Tuple[Array Array]
            Tuple of function value and gradient of shape (n_parameters,).

        """
        # TODO: Fix typing
        states, dg_dp_list = self._propagation.get_value_and_gradient(times=times)  # gradient of states wrt parameters
        overlaps = []
        for ii, s in enumerate(self._target_costates.T):
            overlaps.append(jnp.vdot(s, states[-1][:, ii]))
        f = jnp.average(jnp.asarray(overlaps))

        df_dp_list = []
        for dg_dp in dg_dp_list[-1]:
            gs = []
            for ii, s in enumerate(self._target_costates.T):
                gs.append(jnp.vdot(s, dg_dp[:, ii]))
            g = jnp.average(jnp.asarray(gs))
            df_dp_list.append(jnp.real(f.conj() * g + f * g.conj()))  # chain rule for abs^2

        fid = UnitaryFidelity._fid(jnp.asarray(overlaps))
        return fid, jnp.array(df_dp_list)  # shape scalar, (n_parameters,)

    def set_ideal_gate(self, gate: Array):
        """Compute target states for the L2 norm.

        Parameters
        ----------
        gate : Array
            Target state computation via this gate.

        """
        if self._basis_states is None:
            self._target_costates = gate
        else:
            self._target_costates = self._basis_states @ gate
